# Adapted from the Tensor2Tensor's implementation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Modifications copyright (C) 2019 Texar
# ==============================================================================
"""
Implementation of beam search with penalties.

Adapted from:
    `https://github.com/tensorflow/tensor2tensor/blob/eb048f69c7ea860324122b87cb9caf59c52a27f3/tensor2tensor/utils/beam_search.py`
"""
from typing import Any, Callable, Optional, Tuple, TypeVar, overload

import torch

from texar.torch.utils import map_structure, torch_bool

__all__ = [
    'beam_search',
]

State = TypeVar('State')

# Default value for INF
INF = 1.0 * 1e7


def gather_nd(params: Any, indices: torch.Tensor) -> Any:
    if not isinstance(params, torch.Tensor):
        return params
    assert len(indices.size()) == 3
    orig_size = params.size()
    index = indices[:, :, 1].view(-1) + indices[:, :, 0].view(-1) * orig_size[1]
    ret = torch.index_select(
        params.view(-1, *params.size()[2:]), dim=0, index=index
    )
    ret = ret.view(orig_size[0], indices.size(1), *orig_size[2:])

    return ret


def _merge_beam_dim(tensor: Any) -> Any:
    r"""Reshapes first two dimensions in to single dimension.

    Args:
        tensor: Tensor to reshape of shape `[A, B, ...]`.

    Returns:
        Reshaped tensor of shape `[A * B, ...]`.
    """
    if not isinstance(tensor, torch.Tensor):
        return tensor
    shape = list(tensor.size())
    shape[0] *= shape[1]  # batch -> batch * beam_size
    shape.pop(1)  # Remove beam dim
    return tensor.view(tuple(shape))


def _unmerge_beam_dim(tensor: Any, batch_size: int,
                      beam_size: int) -> Any:
    r"""Reshapes first dimension back to `[batch_size, beam_size]`.

    Args:
        tensor: Tensor to reshape of shape `[batch_size * beam_size, ...]`.
        batch_size: int, original batch size.
        beam_size: int, original beam size.

    Returns:
        Reshaped tensor of shape `[batch_size, beam_size, ...]`.
    """
    if not isinstance(tensor, torch.Tensor):
        return tensor
    shape = list(tensor.size())
    new_shape = [batch_size] + [beam_size] + shape[1:]
    return tensor.view(tuple(new_shape))


def _expand_to_beam_size(tensor: Any,
                         beam_size: int) -> Any:
    r"""Tiles a given tensor by :attr:`beam_size`.

    Args:
        tensor: tensor to tile. Shape: `[batch_size, ...]`.
        beam_size: How much to tile the tensor by.

    Returns:
        Tiled tensor of shape `[batch_size, beam_size, ...]`.
    """
    if not isinstance(tensor, torch.Tensor):
        return tensor
    tensor = torch.unsqueeze(tensor, dim=1)
    tile_dims = [1] * len(tensor.size())
    tile_dims[1] = beam_size

    return tensor.repeat(tuple(tile_dims))


def log_prob_from_logits(logits: torch.Tensor) -> torch.Tensor:
    return logits - torch.logsumexp(logits, dim=-1, keepdim=True)


def compute_batch_indices(batch_size: int, beam_size: int) -> torch.LongTensor:
    r"""Computes the i-th coordinate that contains the batch index for
    gathers.

    The batch index tensor is a tensor like `[[0,0,0,0,],[1,1,1,1],..]`.
    It says which batch the beam item is in. This will create the first
    dimension of the 2D coordinates needed for the gather.

    Args:
        batch_size: Batch size
        beam_size: Size of the beam.

    Returns:
        `[batch_size, beam_size]` tensor of ids.
    """
    batch_pos = torch.arange(batch_size)
    batch_pos = batch_pos.view(-1, 1).expand(batch_size, beam_size)
    return batch_pos


def compute_topk_scores_and_seq(
    sequences: torch.LongTensor,
    scores: torch.Tensor,
    scores_to_gather: torch.Tensor,
    flags: torch.ByteTensor,
    beam_size: int,
    batch_size: int,
    states_to_gather: Optional[State] = None,
) -> Tuple[torch.LongTensor, torch.Tensor, torch.ByteTensor, Optional[State]]:
    r"""Given sequences and scores, will gather the top-k (`k = beam`) size
    sequences.

    This function is used to grow alive, and finished. It takes sequences,
    scores, and flags, and returns the top k from sequence
    :attr:`scores_to_gather`, and flags based on the values in scores.

    Args:
        sequences: Tensor of sequences that we need to gather from.
            Shape: `[batch_size, beam_size, seq_length]`.
        scores: Tensor of scores for each sequence in sequences. We will use
            these to compute the top-k. Shape: `[batch_size, beam_size]`.
        scores_to_gather: Tensor of scores for each sequence in sequences.
            Shape: `[batch_size, beam_size]`.
            We will return the gathered scores from here.
            Scores to gather is different from scores because for
            grow_alive, we will need to return log-probabilities, while for
            grow_finished, we will need to return the length penalized
            scores.
        flags: Tensor of booleans for sequences that say whether a sequence
            has reached `EOS`.
        beam_size: int
        batch_size: int
        states_to_gather: (possibly nested structure of) decoding states.

    :returns: Tuple of:

        - `topk_seq`: `[batch_size, beam_size, decode_length]`.
        - `topk_gathered_scores`: `[batch_size, beam_size]`.
        - `topk_finished_flags`: `[batch_size, beam_size]`.
    """
    # by default top-k is for the last dimension
    _, topk_indexes = torch.topk(scores, k=beam_size)
    # The next three steps are to create coordinates for torch.gather_nd to
    # pull out the top-k sequences from sequences based on scores.
    # batch pos is a tensor like [[0,0,0,0,],[1,1,1,1],..]. It says which
    # batch the beam item is in. This will create the i of the i,j
    # coordinate needed for the gather
    batch_pos = compute_batch_indices(batch_size, beam_size)
    batch_pos = batch_pos.to(device=topk_indexes.device)
    # top coordinates will give us the actual coordinates to do the gather.
    # stacking will create a tensor of dimension batch * beam * 2, where
    # the last dimension contains the i,j gathering coordinates.
    top_coordinates = torch.stack([batch_pos, topk_indexes], dim=2)

    # Gather up the highest scoring sequences.
    topk_seq = gather_nd(sequences, top_coordinates)
    topk_flags = gather_nd(flags, top_coordinates)
    topk_gathered_scores = gather_nd(scores_to_gather, top_coordinates)
    if states_to_gather is not None:
        topk_gathered_states = map_structure(
            lambda state: gather_nd(state, top_coordinates), states_to_gather
        )
    else:
        topk_gathered_states = states_to_gather
    return topk_seq, topk_gathered_scores, topk_flags, topk_gathered_states


# TODO: Remove these once pylint supports function stubs.
# pylint: disable=unused-argument,function-redefined

@overload
def beam_search(
    symbols_to_logits_fn: Callable[[torch.Tensor, State],
                                   Tuple[torch.Tensor, State]],
    initial_ids: torch.LongTensor,
    beam_size: int,
    decode_length: int,
    vocab_size: int,
    alpha: float,
    eos_id: int,
    states: State,
    stop_early: bool = True) -> Tuple[torch.LongTensor, torch.Tensor]: ...


@overload
def beam_search(
    symbols_to_logits_fn: Callable[[torch.Tensor], torch.Tensor],
    initial_ids: torch.LongTensor,
    beam_size: int,
    decode_length: int,
    vocab_size: int,
    alpha: float,
    eos_id: int,
    states: Optional[State] = None,
    stop_early: bool = True) -> Tuple[torch.LongTensor, torch.Tensor]: ...

# pylint: enable=unused-argument


def beam_search(
    symbols_to_logits_fn,
    initial_ids,
    beam_size,
    decode_length,
    vocab_size,
    alpha,
    eos_id,
    states=None,
    stop_early=True,
):
    r"""Beam search with length penalties.

    Requires a function that can take the currently decoded symbols and
    return the logits for the next symbol. The implementation is inspired
    by https://arxiv.org/abs/1609.08144.

    Variables used within this function follow the naming pattern:
    `(alive|finished)_topk_(seq,scores)`.

    Variables marked `alive` represent the new beam sequences that will be
    processed in the next step.    Variables marked `finished` represent
    the completed beam sequences, which may be padded with 0 if no beams
    finished.

    Variables marked `seq` store the full beam sequence for the time step.
    Variables marked `scores` store the sequence's final log scores.

    The beam search steps will be processed sequentially in order, so when
    capturing observed from these operations, tensors, clients can make
    assumptions about which step is being recorded.

    Args:
        symbols_to_logits_fn: Interface to the model, to provide logits.
            Should take `[batch_size, decoded_ids]` and return
            `[batch_size, vocab_size]`.
        initial_ids: LongTensor of shape `[batch_size]`. IDs to start off the
            decoding, this will be the first thing handed to
            :attr:`symbols_to_logits_fn` (after expanding to beam size).
        beam_size: Size of the beam.
        decode_length: Number of steps to decode for.
        vocab_size: Size of the vocab, must equal the size of the logits
            returned by :attr:`symbols_to_logits_fn`.
        alpha: alpha for length penalty.
        eos_id: ID for end of sentence.
        states: (possibly nested structure of) decoding states.
        stop_early: a boolean - stop once best sequence is provably
            determined.

    Returns:
        Tuple of

        - decoded beams (shape: `[batch_size, beam_size, decode_length]`)
        - decoding probabilities (shape: `[batch_size, beam_size]`)
    """

    batch_size = initial_ids.size()[0]

    # Assume initial_ids are prob 1.0
    initial_log_probs = torch.Tensor(
        [[0.0] + [-float("inf")] * (beam_size - 1)]
    )  # [1, beam_size]
    initial_log_probs = initial_log_probs.to(device=initial_ids.device)
    # Expand to beam_size (batch_size, beam_size)
    alive_log_probs = initial_log_probs.repeat((batch_size, 1))

    # Expand each batch and state to beam_size
    alive_seq = _expand_to_beam_size(initial_ids, beam_size)
    alive_seq = torch.unsqueeze(alive_seq, dim=2)
    # (batch_size, beam_size, 1)

    if states is not None:
        states = map_structure(
            lambda state: _expand_to_beam_size(state, beam_size), states
        )

    # Finished will keep track of all the sequences that have finished so
    # far
    # Finished log-probs will be negative infinity in the beginning
    # finished_flags will keep track of booleans
    finished_seq = torch.zeros(alive_seq.size(), dtype=torch.long)
    # Setting the scores of the initial to negative infinity.
    finished_scores = torch.full((batch_size, beam_size), -INF)
    finished_flags = torch.zeros((batch_size, beam_size), dtype=torch_bool)

    finished_seq = finished_seq.to(device=initial_ids.device)
    finished_scores = finished_scores.to(device=initial_ids.device)
    finished_flags = finished_flags.to(device=initial_ids.device)

    def grow_finished(
        finished_seq: torch.LongTensor,
        finished_scores: torch.Tensor,
        finished_flags: torch.ByteTensor,
        curr_seq: torch.LongTensor,
        curr_scores: torch.Tensor,
        curr_finished: torch.ByteTensor,
    ) -> Tuple[torch.LongTensor, torch.Tensor, torch.ByteTensor]:
        r"""Given sequences and scores, will gather the top-k (`k = beam`) size
        sequences.

        Args:
            finished_seq: Finished sequences.
                Shape: `[batch_size, beam_size, current_decoded_length]`.
            finished_scores: Scores for each finished sequences.
                Shape: `[batch_size, beam_size]`.
            finished_flags: Finished flags for each of these sequences.
                Shape: `[batch_size, beam_size]`
            curr_seq: Top-k sequences that has been grown by one
                position.
                Shape: `[batch_size, beam_size, current_decoded_length]`.
            curr_scores: Scores for each of the top-k sequences.
                Shape: `[batch_size, beam_size]`.
            curr_finished: Finished flags for each of the top-k sequences.
                Shape: `[batch_size, beam_size]`.

        Returns:
            Tuple of

            - Top-k sequences based on scores.
            - Log-probabilities of these sequences.
            - Finished flags of these sequences.
        """
        # First append a column of 0'ids to finished to make the same
        # length with finished scores
        _appended = torch.zeros(batch_size, beam_size, 1, dtype=torch.long)
        _appended = _appended.to(device=finished_seq.device)
        finished_seq = torch.cat([finished_seq, _appended], dim=2)

        # Set the scores of the unfinished seq in curr_seq to large
        # negative values
        curr_scores = curr_scores + (1.0 - curr_finished.float()) * -INF
        # concatenating the sequences and scores along beam axis
        curr_finished_seq = torch.cat([finished_seq, curr_seq], dim=1)
        curr_finished_scores = torch.cat([finished_scores, curr_scores], dim=1)
        curr_finished_flags = torch.cat([finished_flags, curr_finished], dim=1)
        next_seq, next_scores, next_flags, _ = compute_topk_scores_and_seq(
            curr_finished_seq,
            curr_finished_scores,
            curr_finished_scores,
            curr_finished_flags,
            beam_size,
            batch_size,
        )
        return next_seq, next_scores, next_flags

    def grow_alive(
        curr_seq: torch.LongTensor, curr_scores: torch.Tensor,
            curr_log_probs: torch.Tensor, curr_finished: torch.ByteTensor,
            states: Optional[State]
    ) -> Tuple[torch.LongTensor, torch.Tensor, torch.ByteTensor,
               Optional[State]]:
        r"""Given sequences and scores, will gather the top k=beam size
        sequences.

        Args:
            curr_seq: Current top-k sequences that has been grown by one
                position.
                Shape: `[batch_size, beam_size, i + 1]`.
            curr_scores: Scores for each of these sequences.
                Shape: `[batch_size, beam_size]`.
            curr_log_probs: Log-probabilities for each of these sequences.
                Shape: `[batch_size, beam_size]`.
            curr_finished: Finished flags for each of these sequences.
                Shape: `[batch_size, beam_size]`.
            states: (possibly nested structure of) decoding states.

        :returns: Tuple of:

            - Top-k sequences based on scores.
            - Log-probabilities of these sequences.
            - Finished flags of these sequences.
            - Decoding states for these sequences.
        """
        # Set the scores of the finished seq in curr_seq to large negative
        # values
        curr_scores = curr_scores + curr_finished.float() * -INF
        return compute_topk_scores_and_seq(
            curr_seq,
            curr_scores,
            curr_log_probs,
            curr_finished,
            beam_size,
            batch_size,
            states,
        )

    def grow_topk(
        i: int, alive_seq: torch.LongTensor, alive_log_probs: torch.Tensor,
        states: Optional[State]
    ) -> Tuple[torch.LongTensor, torch.Tensor, torch.Tensor,
               torch.ByteTensor, Optional[State]]:
        r"""Inner beam search loop.

        This function takes the current alive sequences, and grows them to
        top-k sequences where `k = 2 * beam`. We use `2 * beam` because we could
        have `beam_size` number of sequences that might hit `<EOS>` and there
        will be no alive sequences to continue. With `2 * beam_size`, this
        will not happen. This relies on the assumption the vocab size is >
        beam size. If this is true, we'll have at least `beam_size` non-`<EOS>`
        extensions if we extract the next top `2 * beam` words.
        Length penalty is given by :math:`(5+len(decode)/6) ^ -\alpha`.

        Please refer to https://arxiv.org/abs/1609.08144.

        Args:
            i: loop index
            alive_seq: Top-k sequences decoded so far.
                Shape: `[batch_size, beam_size, i + 1]`.
            alive_log_probs: Log-probabilities of these sequences.
                Shape: `[batch_size, beam_size]`
            states: (possibly nested structure of) decoding states.

        :returns: Tuple of:

            - Top-k sequences extended by the next word.
            - Log-probabilities of these sequences,
            - The scores with length penalty of these sequences,
            - Flags indicating which of these sequences have finished
              decoding.
            - Transformed decoding states with same structure as :attr:`state`.
        """
        # Get the logits for all the possible next symbols
        flat_ids = alive_seq.view(batch_size * beam_size, -1)

        # (batch_size * beam_size, decoded_length)
        if states is not None:
            flat_states = map_structure(_merge_beam_dim, states)
            flat_logits, flat_states = symbols_to_logits_fn(
                flat_ids, flat_states
            )
            states = map_structure(
                lambda t: _unmerge_beam_dim(t, batch_size, beam_size),
                flat_states,
            )
        else:
            flat_logits = symbols_to_logits_fn(flat_ids)
        logits = flat_logits.view(batch_size, beam_size, -1)

        # Convert logits to normalized log-probs
        candidate_log_probs = log_prob_from_logits(logits)

        # Multiply the probabilities by the current probabilities of the
        # beam.
        # (batch_size, beam_size, vocab_size) + (batch_size, beam_size, 1)
        log_probs = candidate_log_probs + alive_log_probs.unsqueeze(dim=2)

        length_penalty = ((5.0 + float(i + 1)) / 6.0) ** alpha

        curr_scores = log_probs / length_penalty
        # Flatten out (beam_size, vocab_size) probs in to a list of
        # possibilities
        flat_curr_scores = curr_scores.view(-1, beam_size * vocab_size)

        topk_scores, topk_ids = torch.topk(flat_curr_scores, k=beam_size * 2)
        # Recovering the log-probs because we will need to send them back
        topk_log_probs = topk_scores * length_penalty

        # Work out what beam the top probabilities are in.
        topk_beam_index = topk_ids / vocab_size
        topk_ids %= vocab_size  # Un-flatten the ids

        # The next three steps are to create coordinates for torch.gather_nd
        # to pull out the correct sequences from id's that we need to grow.
        # We will also use the coordinates to gather the booleans of the
        # beam items that survived.
        batch_pos = compute_batch_indices(batch_size, beam_size * 2)
        batch_pos = batch_pos.to(device=topk_beam_index.device)
        # top beams will give us the actual coordinates to do the gather.
        # stacking will create a tensor of dimension batch * beam * 2,
        # where the last dimension contains the i,j gathering coordinates.
        topk_coordinates = torch.stack([batch_pos, topk_beam_index], dim=2)
        # [batch_size, beam_size, 2]

        topk_seq = gather_nd(alive_seq, topk_coordinates)

        if states is not None:
            states = map_structure(
                lambda state: gather_nd(state, topk_coordinates), states
            )

        # Append the most probable alive
        topk_seq = torch.cat([topk_seq, topk_ids.unsqueeze(dim=2)], dim=2)

        topk_finished = topk_ids == eos_id

        return topk_seq, topk_log_probs, topk_scores, topk_finished, states

    def inner_loop(
        i: int,
        alive_seq: torch.LongTensor,
        alive_log_probs: torch.Tensor,
        finished_seq: torch.LongTensor,
        finished_scores: torch.Tensor,
        finished_flags: torch.ByteTensor,
        states: Optional[State],
    ) -> Tuple[int, torch.LongTensor, torch.Tensor, torch.LongTensor,
               torch.Tensor, torch.ByteTensor, Optional[State]]:
        r"""Inner beam search loop.

        There are three groups of tensors: `alive`, `finished`, and `top-k`.

        - The `alive` group contains information about the current alive
          sequences.
        - The `top-k` group contains information about `alive + top_k`
          current decoded words.
        - The `finished` group contains information about finished sentences,
          that is, the ones that have decoded to `<EOS>`. These are what we
          return.

        The general beam search algorithm is as follows:

            While not terminated (please refer to termination condition):

            1. Grow the current `alive` to get `beam * 2` top-k sequences.
            2. Among the `top-k`, move the top `beam_size` ones that haven't
               reached `EOS` into `alive`.
            3. Among the `top-k`, move the top `beam_size` ones have reached
               `EOS` into `finished`.

            Repeat

        To make things simple with using fixed size tensors, we will end
        up inserting unfinished sequences into finished in the beginning.
        To prevent that we add `-INF` to the score of the unfinished
        sequence so that when a true finished sequence does appear, it
        will have a higher score than all the unfinished ones.

        Args:
            i: Loop index
            alive_seq: Topk sequences decoded so far
                Shape: `[batch_size, beam_size, i + 1]`.
            alive_log_probs: Log-probabilities of the beams.
                Shape: `[batch_size, beam_size]`
            finished_seq: Current finished sequences.
                Shape: `[batch_size, beam_size, i+1]`.
            finished_scores: Scores for each of these sequences.
                Shape: `[batch_size, beam_size]`.
            finished_flags: Finished flags for each of these sequences.
                Shape: `[batch_size, beam_size]`
            states: (possibly nested structure of) decoding states.

        :returns: Tuple of:

            - Incremented loop index.
            - New `alive` sequences.
            - Log-probabilities of the `alive` sequences.
            - New `finished` sequences.
            - Scores of the `finished` sequences.
            - Flags indicating which sequences in `finished` has reached `EOS`.
            - Final decoding states with same structure as :attr:`state`.
        """

        # Each inner loop, we carry out three steps:
        # 1. Get the current top-k items.
        # 2. Extract the ones that have finished and haven't finished
        # 3. Recompute the contents of finished based on scores.
        topk_seq, topk_log_probs, topk_scores, topk_finished, \
                states = grow_topk(i, alive_seq, alive_log_probs, states)

        alive_seq, alive_log_probs, _, states = grow_alive(
            topk_seq, topk_scores, topk_log_probs, topk_finished, states
        )
        finished_seq, finished_scores, finished_flags = grow_finished(
            finished_seq,
            finished_scores,
            finished_flags,
            topk_seq,
            topk_scores,
            topk_finished,
        )

        return (
            i + 1,
            alive_seq,
            alive_log_probs,
            finished_seq,
            finished_scores,
            finished_flags,
            states,
        )

    def _is_finished(
        i: int,
        alive_log_probs: torch.Tensor,
        finished_scores: torch.Tensor
    ) -> bool:
        r"""Check termination condition.

        We terminate when we decoded up to `decode_length` or the lowest
        scoring item in finished has a greater score that the highest probable
        item in alive divided by the max length penalty.

        Args:
            i: Loop index
            alive_log_probs: Log-probabilities of the beams.
                Shape: `[batch_size, beam_size]`.
            finished_scores: Scores for each of these sequences.
                Shape: `[batch_size, beam_size]`.

        Returns:
            Bool.
        """
        max_length_penalty = ((5.0 + float(decode_length)) / 6.0) ** alpha
        # The best possible score of the most likely alive sequence
        lower_bound_alive_scores = alive_log_probs[:, 0] / max_length_penalty

        if not stop_early:
            # by considering the min score (in the top N beams) we ensure that
            # the decoder will keep decoding until there is at least one beam
            # (in the top N) that can be improved (w.r.t. the alive beams).
            # any unfinished beam will have score -INF - thus the min
            # will always be -INF if there is at least one unfinished beam -
            # which means the bound_is_met condition cannot be true in this
            # case.
            lowest_score_of_finished_in_finished = torch.min(finished_scores)
        else:
            # by taking the max score we only care about the first beam;
            # as soon as this first beam cannot be beaten from the alive beams
            # the beam decoder can stop.
            # similarly to the above, if the top beam is not completed, its
            # finished_score is -INF, thus it will not activate the
            # bound_is_met condition. (i.e., decoder will keep going on).
            # note we need to find the max for every sequence eparately - so,
            # we need to keep the batch dimension (see axis=1)
            lowest_score_of_finished_in_finished, _ = torch.max(finished_scores,
                                                                dim=1)

        bound_is_met = (
            (lowest_score_of_finished_in_finished > lower_bound_alive_scores)
            .all()
            .item()
        )

        ret = (i < decode_length) & (~bound_is_met)

        return ret

    step = 0
    while _is_finished(step, alive_log_probs, finished_scores):
        step, alive_seq, alive_log_probs, finished_seq, finished_scores, \
                finished_flags, states = inner_loop(
            step,
            alive_seq,
            alive_log_probs,
            finished_seq,
            finished_scores,
            finished_flags,
            states,
        )

    # Accounting for corner case: It's possible that no sequence in alive
    # for a particular batch item ever reached EOS. In that case, we
    # should just copy the contents of alive for that batch item. tf
    # reduce_any(finished_flags, 1)
    # if 0, means that no sequence for that batch index had reached EOS.
    # We need to do the same for the scores as well.

    ret_seq, ret_scores = [], []
    for idx, flag_per_instance in enumerate(finished_flags.any(dim=1).tolist()):
        if flag_per_instance:
            ret_seq.append(finished_seq[idx])
            ret_scores.append(finished_scores[idx])
        else:
            ret_seq.append(alive_seq[idx])
            ret_scores.append(alive_log_probs[idx])

    ret_seq = torch.stack(ret_seq, dim=0)
    ret_scores = torch.stack(ret_scores, dim=0)

    return ret_seq, ret_scores

# pylint: enable=function-redefined



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pre-trained XLNet Tokenizer.

Code structure adapted from:
    `https://github.com/huggingface/pytorch-transformers/blob/master/pytorch_transformers/tokenization_xlnet.py`
"""

from typing import Any, Dict, List, Optional, Tuple

import os
import unicodedata
from shutil import copyfile
import sentencepiece as spm

from texar.torch.modules.pretrained.xlnet import PretrainedXLNetMixin
from texar.torch.data.tokenizers.tokenizer_base import TokenizerBase
from texar.torch.utils.utils import truncate_seq_pair

__all__ = [
    "XLNetTokenizer",
]

SPIECE_UNDERLINE = u'▁'

SEG_ID_A = 0
SEG_ID_B = 1
SEG_ID_CLS = 2
SEG_ID_SEP = 3
SEG_ID_PAD = 4


class XLNetTokenizer(PretrainedXLNetMixin, TokenizerBase):
    r"""Pre-trained XLNet Tokenizer.

    Args:
        pretrained_model_name (optional): a `str`, the name of
            pre-trained model (e.g., `xlnet-base-uncased`). Please refer to
            :class:`~texar.torch.modules.PretrainedXLNetMixin` for
            all supported models.
            If None, the model name in :attr:`hparams` is used.
        cache_dir (optional): the path to a folder in which the
            pre-trained models will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.
    """

    _IS_PRETRAINED = True
    _MAX_INPUT_SIZE = {
        'xlnet-base-cased': None,
        'xlnet-large-cased': None,
    }
    _VOCAB_FILE_NAMES = {'vocab_file': 'spiece.model'}
    _VOCAB_FILE_MAP = {
        'vocab_file': {
            'xlnet-base-cased': 'spiece.model',
            'xlnet-large-cased': 'spiece.model',
        }
    }

    def __init__(self,
                 pretrained_model_name: Optional[str] = None,
                 cache_dir: Optional[str] = None,
                 hparams=None):
        self.load_pretrained_config(pretrained_model_name, cache_dir, hparams)

        super().__init__(hparams=None)

        self.__dict__: Dict

        self.config = {
            'do_lower_case': self.hparams['do_lower_case'],
            'remove_space': self.hparams['remove_space'],
            'keep_accents': self.hparams['keep_accents'],
        }

        if self.pretrained_model_dir is not None:
            assert self.pretrained_model_name is not None
            vocab_file = os.path.join(self.pretrained_model_dir,
                                      self._VOCAB_FILE_MAP['vocab_file']
                                      [self.pretrained_model_name])
            assert self.pretrained_model_name is not None
            if self._MAX_INPUT_SIZE.get(self.pretrained_model_name):
                self.max_len = self._MAX_INPUT_SIZE[self.pretrained_model_name]
        else:
            vocab_file = self.hparams['vocab_file']
            if self.hparams.get('max_len'):
                self.max_len = self.hparams['max_len']

        if not os.path.isfile(vocab_file):
            raise ValueError("Can't find a vocabulary file at path "
                             "'{}".format(vocab_file))

        self.do_lower_case = self.hparams["do_lower_case"]
        self.remove_space = self.hparams["remove_space"]
        self.keep_accents = self.hparams["keep_accents"]
        self.vocab_file = vocab_file

        self.sp_model = spm.SentencePieceProcessor()
        self.sp_model.Load(self.vocab_file)

    # spm.SentencePieceProcessor() is a SwigPyObject object which cannot be
    # pickled. We need to define __getstate__ here.
    def __getstate__(self):
        state = self.__dict__.copy()
        state["sp_model"] = None
        state["vocab_file"] = None
        return state, self.vocab_file

    # spm.SentencePieceProcessor() is a SwigPyObject object which cannot be
    # pickled. We need to define __setstate__ here.
    def __setstate__(self, d):
        self.__dict__, self.vocab_file = d
        self.sp_model = spm.SentencePieceProcessor()
        self.sp_model.Load(self.vocab_file)

    def _preprocess_text(self, inputs: str) -> str:
        r"""Pre-process the text, including removing space,
        stripping accents, and lower-casing the text.
        """
        if self.remove_space:
            outputs = ' '.join(inputs.strip().split())
        else:
            outputs = inputs
        outputs = outputs.replace("``", '"').replace("''", '"')

        if not self.keep_accents:
            outputs = unicodedata.normalize('NFKD', outputs)
            outputs = ''.join([c for c in outputs if not
            unicodedata.combining(c)])
        if self.do_lower_case:
            outputs = outputs.lower()

        return outputs

    def _map_text_to_token(self, text: str,  # type: ignore
                           sample: bool = False) -> List[str]:
        text = self._preprocess_text(text)
        if not sample:
            pieces = self.sp_model.EncodeAsPieces(text)
        else:
            pieces = self.sp_model.SampleEncodeAsPieces(text, 64, 0.1)

        new_pieces: List[str] = []
        for piece in pieces:
            if len(piece) > 1 and piece[-1] == ',' and piece[-2].isdigit():
                cur_pieces = self.sp_model.EncodeAsPieces(
                    piece[:-1].replace(SPIECE_UNDERLINE, ''))
                if piece[0] != SPIECE_UNDERLINE and \
                        cur_pieces[0][0] == SPIECE_UNDERLINE:
                    if len(cur_pieces[0]) == 1:
                        cur_pieces = cur_pieces[1:]
                    else:
                        cur_pieces[0] = cur_pieces[0][1:]
                cur_pieces.append(piece[-1])
                new_pieces.extend(cur_pieces)
            else:
                new_pieces.append(piece)

        return new_pieces

    def save_vocab(self, save_dir: str) -> Tuple[str]:
        r"""Save the sentencepiece vocabulary (copy original file) to
        a directory.
        """
        if not os.path.isdir(save_dir):
            raise ValueError("Vocabulary path ({}) should be a "
                             "directory".format(save_dir))
        out_vocab_file = os.path.join(save_dir,
                                      self._VOCAB_FILE_NAMES['vocab_file'])

        if os.path.abspath(self.vocab_file) != os.path.abspath(out_vocab_file):
            copyfile(self.vocab_file, out_vocab_file)

        return (out_vocab_file,)

    @property
    def vocab_size(self) -> int:
        return len(self.sp_model)

    def _map_token_to_id(self, token: str) -> int:
        return self.sp_model.PieceToId(token)

    def _map_id_to_token(self, index: int) -> str:
        token = self.sp_model.IdToPiece(index)
        return token

    def map_token_to_text(self, tokens: List[str]) -> str:
        r"""Maps a sequence of tokens (string) in a single string."""
        out_string = ''.join(tokens).replace(SPIECE_UNDERLINE, ' ').strip()
        return out_string

    def encode_text(self,
                    text_a: str,
                    text_b: Optional[str] = None,
                    max_seq_length: Optional[int] = None) -> \
            Tuple[List[int], List[int], List[int]]:
        r"""Adds special tokens to a sequence or sequence pair and computes the
        corresponding segment ids and input mask for XLNet specific tasks.
        The sequence will be truncated if its length is larger than
        ``max_seq_length``.

        A XLNet sequence has the following format:
        X `[sep_token]` `[cls_token]`

        A XLNet sequence pair has the following format:
        `[cls_token]` A `[sep_token]` B `[sep_token]`

        Args:
            text_a: The first input text.
            text_b: The second input text.
            max_seq_length: Maximum sequence length.

        Returns:
            A tuple of `(input_ids, segment_ids, input_mask)`, where

            - ``input_ids``: A list of input token ids with added
              special token ids.
            - ``segment_ids``: A list of segment ids.
            - ``input_mask``: A list of mask ids. The mask has 1 for real
              tokens and 0 for padding tokens. Only real tokens are
              attended to.
        """
        if max_seq_length is None:
            max_seq_length = self.max_len

        cls_token_id = self._map_token_to_id(self.cls_token)
        sep_token_id = self._map_token_to_id(self.sep_token)

        token_ids_a = self.map_text_to_id(text_a)
        assert isinstance(token_ids_a, list)

        token_ids_b = None
        if text_b:
            token_ids_b = self.map_text_to_id(text_b)

        if token_ids_b:
            assert isinstance(token_ids_b, list)
            # Modifies `token_ids_a` and `token_ids_b` in place so that the
            # total length is less than the specified length.
            # Account for [CLS], [SEP], [SEP] with "- 3"
            truncate_seq_pair(token_ids_a, token_ids_b, max_seq_length - 3)

            input_ids = (token_ids_a + [sep_token_id] + token_ids_b +
                         [sep_token_id] + [cls_token_id])
            segment_ids = [SEG_ID_A] * (len(token_ids_a) + 1) + \
                          [SEG_ID_B] * (len(token_ids_b) + 1) + [SEG_ID_CLS]
        else:
            # Account for [CLS] and [SEP] with "- 2"
            token_ids = token_ids_a[:max_seq_length - 2]

            input_ids = token_ids + [sep_token_id] + [cls_token_id]
            segment_ids = [SEG_ID_A] * (len(input_ids) - 1) + [SEG_ID_CLS]

        input_mask = [0] * len(input_ids)

        # Zero-pad up to the maximum sequence length.
        input_ids = [0] * (max_seq_length - len(input_ids)) + input_ids
        input_mask = [1] * (max_seq_length - len(input_mask)) + input_mask
        segment_ids = ([SEG_ID_PAD] * (max_seq_length - len(segment_ids)) +
                       segment_ids)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length

        return input_ids, segment_ids, input_mask

    def encode_text_for_generation(
            self,
            text: str,
            max_seq_length: Optional[int] = None,
            append_eos_token: bool = True) -> Tuple[List[int], int]:
        r"""Adds special tokens to a sequence and computes the corresponding
        sequence length for XLNet specific tasks. The sequence will be truncated
        if its length is larger than ``max_seq_length``.

        A XLNet sequence has the following format:
        `[bos_token]` X `[eos_token]` `[pad_token]`

        Args:
            text: Input text.
            max_seq_length: Maximum sequence length.
            append_eos_token: Whether to append ``eos_token`` after the
                sequence.

        Returns:
            A tuple of `(input_ids, seq_len)`, where

            - ``input_ids``: A list of input token ids with added
              special tokens.
            - ``seq_len``: The sequence length.
        """
        if max_seq_length is None:
            max_seq_length = self.max_len

        token_ids = self.map_text_to_id(text)
        assert isinstance(token_ids, list)

        bos_token_id = self._map_token_to_id(self.bos_token)
        eos_token_id = self._map_token_to_id(self.eos_token)
        pad_token_id = self._map_token_to_id(self.pad_token)

        if append_eos_token:
            input_ids = token_ids[:max_seq_length - 2]
            input_ids = [bos_token_id] + input_ids + [eos_token_id]
        else:
            input_ids = token_ids[:max_seq_length - 1]
            input_ids = [bos_token_id] + input_ids

        seq_len = len(input_ids)

        # Pad up to the maximum sequence length.
        input_ids = input_ids + [pad_token_id] * (max_seq_length - seq_len)

        assert len(input_ids) == max_seq_length

        return input_ids, seq_len

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        * The tokenizer is determined by the constructor argument
          :attr:`pretrained_model_name` if it's specified. In this case,
          `hparams` are ignored.
        * Otherwise, the tokenizer is determined by
          `hparams['pretrained_model_name']` if it's specified. All other
          configurations in `hparams` are ignored.
        * If the above two are `None`, the tokenizer is defined by the
          configurations in `hparams`.

        .. code-block:: python

            {
                "pretrained_model_name": "xlnet-base-cased",
                "vocab_file": None,
                "max_len": None,
                "bos_token": "<s>",
                "eos_token": "</s>",
                "unk_token": "<unk>",
                "sep_token": "<sep>",
                "pad_token": "<pad>",
                "cls_token": "<cls>",
                "mask_token": "<mask>",
                "additional_special_tokens": ["<eop>", "<eod>"],
                "do_lower_case": False,
                "remove_space": True,
                "keep_accents": False,
                "name": "xlnet_tokenizer",
            }

        Here:

        `"pretrained_model_name"`: str or None
            The name of the pre-trained XLNet model.

        `"vocab_file"`: str or None
            The path to a sentencepiece vocabulary file.

        `"max_len"`: int or None
            The maximum sequence length that this model might ever be used with.

        `"bos_token"`: str
            Beginning of sentence token.

        `"eos_token"`: str
            End of sentence token.

        `"unk_token"`: str
            Unknown token.

        `"sep_token"`: str
            Separation token.

        `"pad_token"`: str
            Padding token.

        `"cls_token"`: str
            Classification token.

        `"mask_token"`: str
            Masking token.

        `"additional_special_tokens"`: list
            A list of additional special tokens.

        `"do_lower_case"`: bool
            Whether to lower-case the text.

        `"remove_space"`: bool
            Whether to remove the space in the text.

        `"keep_accents"`: bool
            Whether to keep the accents in the text.

        `"name"`: str
            Name of the tokenizer.
        """
        return {
            'pretrained_model_name': 'xlnet-base-cased',
            'vocab_file': None,
            'max_len': None,
            'bos_token': '<s>',
            'eos_token': '</s>',
            'unk_token': '<unk>',
            'sep_token': '<sep>',
            'pad_token': '<pad>',
            'cls_token': '<cls>',
            'mask_token': '<mask>',
            'additional_special_tokens': ['<eop>', '<eod>'],
            'do_lower_case': False,
            'remove_space': True,
            'keep_accents': False,
            'name': 'xlnet_tokenizer',
            '@no_typecheck': ['pretrained_model_name'],
        }

    @classmethod
    def _transform_config(cls, pretrained_model_name: str,
                          cache_dir: str):
        r"""Returns the configuration of the pre-trained XLNet tokenizer."""
        return {
            'vocab_file': None,
            'max_len': None,
            'bos_token': '<s>',
            'eos_token': '</s>',
            'unk_token': '<unk>',
            'sep_token': '<sep>',
            'pad_token': '<pad>',
            'cls_token': '<cls>',
            'mask_token': '<mask>',
            'additional_special_tokens': ['<eop>', '<eod>'],
            'do_lower_case': False,
            'remove_space': True,
            'keep_accents': False,
        }



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Miscellaneous Utility functions.
"""

import collections
import copy
import inspect
from functools import lru_cache
from pydoc import locate
from typing import (
    Any, Callable, Collection, Dict, List, MutableMapping, Optional, Sequence,
    Tuple, Type, TypeVar, Union, cast, no_type_check, overload)

import funcsigs
import numpy as np
import torch
import torch.nn as nn
from torch.nn.modules.conv import _ConvNd

from texar.torch.hyperparams import HParams
from texar.torch.utils.dtypes import _maybe_list_to_array
from texar.torch.utils.types import MaybeSeq, MaybeTuple

MAX_SEQ_LENGTH = np.iinfo(np.int32).max

__all__ = [
    'no_map',
    'map_structure',
    'map_structure_zip',
    'get_first_in_structure',
    'sequence_mask',
    'get_args',
    'get_default_arg_values',
    'check_or_get_class',
    'get_class',
    'check_or_get_instance',
    'get_instance',
    'check_or_get_instance_with_redundant_kwargs',
    'get_instance_with_redundant_kwargs',
    'get_function',
    'call_function_with_redundant_kwargs',
    'get_instance_kwargs',
    'dict_patch',
    'dict_lookup',
    'dict_fetch',
    'dict_pop',
    'flatten_dict',
    'strip_token',
    'strip_eos',
    'strip_bos',
    'strip_special_tokens',
    'str_join',
    'default_str',
    'uniquify_str',
    'ceildiv',
    'sum_tensors',
]

T = TypeVar('T')  # type argument
R = TypeVar('R')  # return type
K = TypeVar('K')  # key type
V = TypeVar('V')  # value type
Kwargs = Dict[str, Any]
AnyDict = MutableMapping[str, Any]
ParamDict = Union[HParams, AnyDict]

Type_size_lambda_map = {
    nn.Linear: lambda x: x.out_features,
    nn.Bilinear: lambda x: x.out_features,
    _ConvNd: lambda x: x.out_channels * len(x.kernel_size),
    nn.Embedding: lambda x: x.embedding_dim,
    nn.EmbeddingBag: lambda x: x.embedding_dim,
    nn.RNNCellBase: lambda x: x.hidden_size,
}

Type_size_keeper = [
    nn.ELU, nn.Hardshrink, nn.Hardtanh, nn.LeakyReLU, nn.LogSigmoid,
    nn.PReLU, nn.ReLU, nn.RReLU, nn.SELU, nn.CELU, nn.Sigmoid, nn.Softplus,
    nn.Softshrink, nn.Softsign, nn.Tanh, nn.Tanhshrink, nn.Threshold,
    nn.Softmin, nn.Softmax, nn.LogSoftmax, nn.Dropout, nn.AlphaDropout,
]

# NestedCollection = Union[T, Collection['NestedCollection']]


@lru_cache(maxsize=None)
def _no_map_type(container_type: Type[T]) -> Type[T]:
    # Create a subtype of the container type that sets an normally inaccessible
    # special attribute on instances.
    # This is necessary because `setattr` does not work on built-in types
    # (e.g. `list`).
    new_type = type("_no_map" + container_type.__name__,
                    (container_type,), {'--no-map--': True})
    return new_type


def no_map(container_type: Type[T], *args, **kwargs) -> T:
    r"""Create a "`non-mappable`" container type, i.e. it will be treated as a
    singleton object in :meth:`map_structure` and :meth:`map_structure_zip`,
    its contents will not be traversed.

    This is implemented by dynamically creating a subclass of the required type,
    and overriding the :attr:`__subclasscheck__` class method to always return
    `False`.

    Args:
        container_type: The type of the container to create,
            e.g. `list`, `dict`.
        args: Arguments to the constructor.
        kwargs: Keyword arguments to the constructor

    Returns:
        The `non-mappable` container type.
    """
    return _no_map_type(container_type)(*args, **kwargs)  # type: ignore


@no_type_check
def map_structure(fn: Callable[[T], R], obj: Collection[T]) -> Collection[R]:
    r"""Map a function over all elements in a (possibly nested) collection.

    Args:
        fn (callable): The function to call on elements.
        obj: The collection to map function over.

    Returns:
        The collection in the same structure, with elements mapped.
    """
    if hasattr(obj, "--no-map--"):
        return fn(obj)
    if isinstance(obj, list):
        return [map_structure(fn, x) for x in obj]
    if isinstance(obj, tuple):
        if isinstance(obj, torch.Size):
            return fn(obj)
        if hasattr(obj, '_fields'):  # namedtuple
            return type(obj)(*[map_structure(fn, x) for x in obj])
        else:
            return tuple(map_structure(fn, x) for x in obj)
    if isinstance(obj, dict):
        return {k: map_structure(fn, v) for k, v in obj.items()}
    if isinstance(obj, set):
        return {map_structure(fn, x) for x in obj}
    return fn(obj)


@no_type_check
def map_structure_zip(fn: Callable[..., R],
                      objs: Sequence[Collection[T]]) -> Collection[R]:
    r"""Map a function over tuples formed by taking one elements from each
    (possibly nested) collection. Each collection must have identical
    structures.

    .. note::
        Although identical structures are required, it is not enforced by
        assertions. The structure of the first collection is assumed to be
        the structure for all collections.

        For rare cases where collections need to have different structures,
        refer to :meth:`no_map`.

    Args:
        fn (callable): The function to call on elements.
        objs: The list of collections to map function over.

    Returns:
        A collection with the same structure, with elements mapped.
    """
    obj = objs[0]
    if hasattr(obj, "--no-map--"):
        return fn(*objs)
    if isinstance(obj, list):
        return [map_structure_zip(fn, xs) for xs in zip(*objs)]
    if isinstance(obj, tuple):
        if isinstance(obj, torch.Size):
            return fn(obj)
        if hasattr(obj, '_fields'):  # namedtuple
            return type(obj)(*[map_structure_zip(fn, xs) for xs in zip(*objs)])
        else:
            return tuple(map_structure_zip(fn, xs) for xs in zip(*objs))
    if isinstance(obj, dict):
        return {k: map_structure_zip(fn, [o[k] for o in objs])
                for k in obj.keys()}
    if isinstance(obj, set):
        return {map_structure_zip(fn, xs) for xs in zip(*objs)}
    return fn(*objs)


def get_first_in_structure(obj: Collection[T]) -> Optional[T]:
    r"""Return the first not-`None` element within a (possibly nested)
    collection.

    Args:
        obj: The collection to pick the element from.

    Returns:
        The first non-`None` element from the collection, or `None` if the
        collection is empty or contains only `None` elements.
    """
    item = None

    def _get_first(x):
        nonlocal item
        if item is None:
            item = x

    map_structure(_get_first, obj)
    return item


def sequence_mask(lengths: Union[torch.LongTensor, List[int]],
                  max_len: Optional[int] = None,
                  dtype: Optional[torch.dtype] = None,
                  device: Optional[torch.device] = None) -> torch.ByteTensor:
    r"""Return a mask tensor representing the first N positions of each cell.

    If ``lengths`` has shape ``[d_1, d_2, ..., d_n]`` the resulting tensor
    ``mask`` has dtype ``dtype`` and shape ``[d_1, d_2, ..., d_n, maxlen]``,
    with

    ```
    mask[i_1, i_2, ..., i_n, j] = (j < lengths[i_1, i_2, ..., i_n])
    ```

    Examples:

    ```python
    sequence_mask([1, 3, 2], 5)  # [[True, False, False, False, False],
                                 #  [True,  True,  True, False, False],
                                 #  [True,  True, False, False, False]]

    sequence_mask([[1, 3],[2,0]])  # [[[ True, False, False],
                                   #   [ True,  True,  True]],
                                   #  [[ True,  True, False],
                                   #   [False, False, False]]]
    ```

    Args:
        lengths: integer tensor or list of int, all its values <= max_len.
        max_len: scalar integer tensor, size of last dimension of returned
            tensor. Default is the maximum value in ``lengths``.
        dtype: the desired data type of returned tensor. Default: if None,
            returns :torch:`ByteTensor`.
        device: the desired device of returned tensor. Default: if None, uses
            the current device for the default tensor type.
    Returns:
        A mask tensor of shape :python:`lengths.shape + (max_len,)`, cast to
        specified dtype.
    Raises:
        ValueError: if ``max_len`` is not a scalar.
    """
    if not isinstance(lengths, torch.Tensor):
        lengths = torch.tensor(lengths, device=device)
    elif device is None:
        device = lengths.device
    lengths: torch.LongTensor
    if max_len is None:
        max_len = torch.max(lengths).item()

    size = lengths.size()
    row_vector = torch.arange(max_len, device=device, dtype=lengths.dtype).view(
        *([1] * len(size)), -1).expand(*size, max_len)
    mask = (row_vector < lengths.unsqueeze(-1)).to(device=device)
    if dtype is not None:
        mask = mask.to(dtype=dtype)

    return mask


def get_args(fn: Callable) -> List[str]:
    r"""Gets the arguments of a function.

    Args:
        fn (callable): The function to inspect.

    Returns:
        list: A list of argument names (``str``) of the function.
    """
    argspec = inspect.getfullargspec(fn)
    args = argspec.args

    # Empty args can be because `fn` is decorated. Use `funcsigs.signature`
    # to re-do the inspect
    if len(args) == 0:
        args = funcsigs.signature(fn).parameters.keys()
        args = list(args)

    return args


def get_default_arg_values(fn: Callable) -> Dict[str, Any]:
    r"""Gets the arguments and respective default values of a function.

    Only arguments with default values are included in the output dictionary.

    Args:
        fn (callable): The function to inspect.

    Returns:
        dict: A dictionary that maps argument names (``str``) to their default
        values. The dictionary is empty if no arguments have default values.
    """
    argspec = inspect.getfullargspec(fn)
    if argspec.defaults is None:
        return {}
    num_defaults = len(argspec.defaults)
    return dict(zip(argspec.args[-num_defaults:], argspec.defaults))


def check_or_get_class(class_or_name: Union[Type[T], str],
                       module_paths: Optional[List[str]] = None,
                       superclass: Optional[MaybeTuple[type]] = None) \
        -> Type[T]:
    r"""Returns the class and checks if the class inherits :attr:`superclass`.

    Args:
        class_or_name: Name or full path to the class, or the class itself.
        module_paths (list, optional): Paths to candidate modules to search
            for the class. This is used if :attr:`class_or_name` is a string and
            the class cannot be located solely based on :attr:`class_or_name`.
            The first module in the list that contains the class
            is used.
        superclass (optional): A (list of) classes that the target class
            must inherit.

    Returns:
        The target class.

    Raises:
        ValueError: If class is not found based on :attr:`class_or_name` and
            :attr:`module_paths`.
        TypeError: If class does not inherits :attr:`superclass`.
    """
    class_ = class_or_name
    if isinstance(class_, str):
        class_ = get_class(class_, module_paths)
    if superclass is not None:
        if not issubclass(class_, superclass):
            raise TypeError(
                "A subclass of {} is expected. Got: {}".format(
                    superclass, class_))
    return class_


def get_class(class_name: str,
              module_paths: Optional[List[str]] = None) -> type:
    r"""Returns the class based on class name.

    Args:
        class_name (str): Name or full path to the class.
        module_paths (list): Paths to candidate modules to search for the
            class. This is used if the class cannot be located solely based on
            ``class_name``. The first module in the list that contains the class
            is used.

    Returns:
        The target class.

    Raises:
        ValueError: If class is not found based on :attr:`class_name` and
            :attr:`module_paths`.
    """
    class_ = locate(class_name)
    if (class_ is None) and (module_paths is not None):
        for module_path in module_paths:
            class_ = locate('.'.join([module_path, class_name]))
            if class_ is not None:
                break

    if class_ is None:
        raise ValueError(
            "Class not found in {}: {}".format(module_paths, class_name))

    return class_  # type: ignore


def check_or_get_instance(ins_or_class_or_name: Union[Type[T], T, str],
                          kwargs: Kwargs,
                          module_paths: Optional[List[str]] = None,
                          classtype: Optional[MaybeTuple[type]] = None) -> T:
    r"""Returns a class instance and checks types.

    Args:
        ins_or_class_or_name: Can be of 3 types:

            - A class to instantiate.
            - A string of the name or full path to a class to instantiate.
            - The class instance to check types.

        kwargs (dict): Keyword arguments for the class constructor. Ignored
            if ``ins_or_class_or_name`` is a class instance.
        module_paths (list, optional): Paths to candidate modules to
            search for the class. This is used if the class cannot be
            located solely based on :attr:`class_name`. The first module
            in the list that contains the class is used.
        classtype (optional): A (list of) class of which the instance must
            be an instantiation.

    Raises:
        ValueError: If class is not found based on :attr:`class_name` and
            :attr:`module_paths`.
        ValueError: If :attr:`kwargs` contains arguments that are invalid
            for the class construction.
        TypeError: If the instance is not an instantiation of
            :attr:`classtype`.
    """
    ret = ins_or_class_or_name
    if isinstance(ret, (str, type)):
        ret = get_instance(ret, kwargs, module_paths)
    if classtype is not None:
        if not isinstance(ret, classtype):
            raise TypeError(
                "An instance of {} is expected. Got: {}".format(classtype, ret))
    return ret


def get_instance(class_or_name: Union[Type[T], str], kwargs: Optional[Kwargs],
                 module_paths: Optional[List[str]] = None) -> T:
    r"""Creates a class instance.

    Args:
        class_or_name: A class, or its name or full path to a class to
            instantiate.
        kwargs (dict): Keyword arguments for the class constructor.
        module_paths (list, optional): Paths to candidate modules to
            search for the class. This is used if the class cannot be
            located solely based on :attr:`class_name`. The first module
            in the list that contains the class is used.

    Returns:
        A class instance.

    Raises:
        ValueError: If class is not found based on :attr:`class_or_name` and
            :attr:`module_paths`.
        ValueError: If :attr:`kwargs` contains arguments that are invalid
            for the class construction.
    """
    # Locate the class
    class_ = class_or_name
    if isinstance(class_, str):
        class_ = get_class(class_, module_paths)

    # Check validity of arguments
    class_args = set(get_args(class_.__init__))

    if kwargs is None:
        kwargs = {}
    for key in kwargs.keys():
        if key not in class_args:
            raise ValueError(
                "Invalid argument for class %s.%s: %s, valid args: %s" %
                (class_.__module__, class_.__name__, key, list(class_args)))

    return class_(**kwargs)  # type: ignore


def check_or_get_instance_with_redundant_kwargs(
        ins_or_class_or_name: Union[Type[T], T, str], kwargs: Kwargs,
        module_paths: Optional[List[str]] = None,
        classtype: Optional[Type[T]] = None) -> T:
    r"""Returns a class instance and checks types.

    Only those keyword arguments in :attr:`kwargs` that are included in the
    class construction method are used.

    Args:
        ins_or_class_or_name: Can be of 3 types:

            - A class to instantiate.
            - A string of the name or module path to a class to instantiate.
            - The class instance to check types.

        kwargs (dict): Keyword arguments for the class constructor.
        module_paths (list, optional): Paths to candidate modules to
            search for the class. This is used if the class cannot be
            located solely based on :attr:`class_name`. The first module
            in the list that contains the class is used.
        classtype (optional): A (list of) classes of which the instance must
            be an instantiation.

    Raises:
        ValueError: If class is not found based on :attr:`class_name` and
            :attr:`module_paths`.
        ValueError: If :attr:`kwargs` contains arguments that are invalid
            for the class construction.
        TypeError: If the instance is not an instantiation of
            :attr:`classtype`.
    """
    ret = ins_or_class_or_name
    if isinstance(ret, (str, type)):
        ret = get_instance_with_redundant_kwargs(ret, kwargs, module_paths)
    if classtype is not None:
        if not isinstance(ret, classtype):
            raise TypeError(
                "An instance of {} is expected. Got: {}".format(classtype, ret))
    return ret


def get_instance_with_redundant_kwargs(
        class_name: Union[Type[T], str], kwargs: Kwargs,
        module_paths: Optional[List[str]] = None) -> T:
    r"""Creates a class instance.

    Only those keyword arguments in :attr:`kwargs` that are included in the
    class construction method are used.

    Args:
        class_name (str): A class or its name or module path.
        kwargs (dict): A dictionary of arguments for the class constructor. It
            may include invalid arguments which will be ignored.
        module_paths (list of str): A list of paths to candidate modules to
            search for the class. This is used if the class cannot be located
            solely based on :attr:`class_name`. The first module in the list
            that contains the class is used.

    Returns:
        A class instance.

    Raises:
        ValueError: If class is not found based on :attr:`class_name` and
            :attr:`module_paths`.
    """
    # Locate the class
    if isinstance(class_name, str):
        class_ = get_class(class_name, module_paths)
    else:
        class_ = class_name

    # Select valid arguments
    selected_kwargs = {}
    class_args = set(get_args(class_.__init__))  # type: ignore
    if kwargs is None:
        kwargs = {}
    for key, value in kwargs.items():
        if key in class_args:
            selected_kwargs[key] = value

    return class_(**selected_kwargs)


def get_function(fn_or_name: Union[str, Callable[[torch.Tensor], torch.Tensor]],
                 module_paths: Optional[List[str]] = None) \
        -> Callable[[torch.Tensor], torch.Tensor]:
    r"""Returns the function of specified name and module.

    Args:
        fn_or_name (str or callable): Name or full path to a function, or the
            function itself.
        module_paths (list, optional): A list of paths to candidate modules to
            search for the function. This is used only when the function
            cannot be located solely based on :attr:`fn_or_name`. The first
            module in the list that contains the function is used.

    Returns:
        A function.

    Raises:
        ValueError: If method with name as :attr:`fn_or_name` is not found.
    """
    if callable(fn_or_name):
        return fn_or_name

    fn = locate(fn_or_name)
    if (fn is None) and (module_paths is not None):
        for module_path in module_paths:
            fn = locate('.'.join([module_path, fn_or_name]))
            if fn is not None:
                break

    if fn is None:
        raise ValueError(
            "Method not found in {}: {}".format(module_paths, fn_or_name))

    return fn  # type: ignore


def call_function_with_redundant_kwargs(fn: Callable[..., R],
                                        kwargs: Dict[str, Any]) -> R:
    r"""Calls a function and returns the results.

    Only those keyword arguments in :attr:`kwargs` that are included in the
    function's argument list are used to call the function.

    Args:
        fn (function): A callable. If :attr:`fn` is not a python function,
            :attr:`fn.__call__` is called.
        kwargs (dict): A ``dict`` of arguments for the callable. It
            may include invalid arguments which will be ignored.

    Returns:
        The returned results by calling :attr:`fn`.
    """
    try:
        fn_args = set(get_args(fn))
    except TypeError:
        fn_args = set(get_args(fn.__call__))  # type: ignore

    if kwargs is None:
        kwargs = {}

    # Select valid arguments
    selected_kwargs = {}
    for key, value in kwargs.items():
        if key in fn_args:
            selected_kwargs[key] = value

    return fn(**selected_kwargs)


def get_instance_kwargs(kwargs: Kwargs, hparams: ParamDict) -> Kwargs:
    r"""Makes a dictionary of keyword arguments with the following structure:

    ``kwargs_ = {'hparams': dict(hparams), **kwargs}``.

    This is typically used for constructing a module which takes a set of
    arguments as well as a argument named ``"hparams"``.

    Args:
        kwargs (dict): A ``dict`` of keyword arguments. Can be `None`.
        hparams: A ``dict`` or an instance of :class:`~texar.torch.HParams`.
            Can be `None`.

    Returns:
        A ``dict`` that contains the keyword arguments in :attr:`kwargs`, and
        an additional keyword argument named ``"hparams"``.
    """
    if hparams is None or isinstance(hparams, dict):
        kwargs_ = {'hparams': hparams}
    elif isinstance(hparams, HParams):
        kwargs_ = {'hparams': hparams.todict()}
    else:
        raise ValueError(
            '`hparams` must be a dict, an instance of HParams, or a `None`.')
    kwargs_.update(kwargs or {})
    return kwargs_


def get_output_size(input_instance: nn.Module) -> Optional[int]:
    r"""Return the final dimension size of :attr:`input_instance` output.

    If type of :attr:`input_instance` is among the common types, the final
    dimension size will be computed.

    Args:
        input_instance: A :class:`~torch.nn.Module` instance from
            which to compute the final dimension size.

    Returns:
        int (optional): The final dimension size of the output.
            If output size is determined by input, returns ``-1``,
            otherwise if output size is not computable, return `None`.
    """

    for t, l in Type_size_lambda_map.items():
        if isinstance(input_instance, t):
            return l(input_instance)
    for t in Type_size_keeper:
        if isinstance(input_instance, t):
            return -1
    return None


def dict_patch(tgt_dict: AnyDict, src_dict: AnyDict) -> AnyDict:
    r"""Recursively patch :attr:`tgt_dict` by adding items from :attr:`src_dict`
    that do not exist in :attr:`tgt_dict`.

    If respective items in :attr:`src_dict` and :attr:`tgt_dict` are both
    ``dict``, the :attr:`tgt_dict` item is patched recursively.

    Args:
        tgt_dict (dict): Target dictionary to patch.
        src_dict (dict): Source dictionary.

    Returns:
        dict: The new :attr:`tgt_dict` that is patched.
    """
    if src_dict is None:
        return tgt_dict

    for key, value in src_dict.items():
        if key not in tgt_dict:
            tgt_dict[key] = copy.deepcopy(value)
        elif isinstance(value, dict) and isinstance(tgt_dict[key], dict):
            tgt_dict[key] = dict_patch(tgt_dict[key], value)
    return tgt_dict


def dict_lookup(dict_: MutableMapping[K, V], keys: Union[List[K], np.ndarray],
                default: Optional[V] = None) -> np.ndarray:
    r"""Looks up :attr:`keys` in the dictionary, returns the corresponding
    values.

    The :attr:`default` is used for keys not present in the dictionary.

    Args:
        dict\_ (dict): A dictionary for lookup.
        keys: A numpy array or a (possibly nested) list of keys.
        default (optional): Value to be returned when a key is not in
            :attr:`dict_`. Error is raised if :attr:`default` is not given and
            key is not in the dictionary.

    Returns:
        A numpy array of values with the same structure as :attr:`keys`.

    Raises:
        TypeError: If key is not in :attr:`dict_` and :attr:`default` is
            `None`.
    """
    return np.vectorize(lambda x: dict_.get(x, default))(keys)  # type: ignore


# TODO: Remove these once pylint supports function stubs.
# pylint: disable=unused-argument,function-redefined,multiple-statements

@overload
def dict_fetch(src_dict: ParamDict,
               tgt_dict_or_keys: Union[ParamDict, List[str]]) -> AnyDict: ...


@overload
def dict_fetch(src_dict: None, tgt_dict_or_keys: Any) -> None: ...


def dict_fetch(src_dict: Optional[ParamDict],
               tgt_dict_or_keys: Union[ParamDict, List[str]]) \
        -> Optional[AnyDict]:
    r"""Fetches a sub-dictionary of :attr:`src_dict` with the keys in
    :attr:`tgt_dict_or_keys`.

    Args:
        src_dict: A dictionary or instance of :class:`~texar.torch.HParams`.
            The source dictionary to fetch values from.
        tgt_dict_or_keys: A dictionary, instance of
            :class:`~texar.torch.HParams`, or a list (or a
            ``dict_keys``/``KeysView``) of keys to be included in the output
            dictionary.

    Returns:
        A new dictionary that is a sub-dictionary of :attr:`src_dict`.
    """
    if src_dict is None:
        return src_dict

    if isinstance(tgt_dict_or_keys, HParams):
        tgt_dict_or_keys = tgt_dict_or_keys.todict()
    if isinstance(tgt_dict_or_keys, MutableMapping):
        tgt_dict_or_keys = tgt_dict_or_keys.keys()  # type: ignore
    keys = list(tgt_dict_or_keys)

    if isinstance(src_dict, HParams):
        src_dict = src_dict.todict()

    return {k: src_dict[k] for k in keys if k in src_dict}


# pylint: enable=unused-argument,function-redefined,multiple-statements


def dict_pop(dict_: MutableMapping[T, Any], pop_keys: MaybeSeq[T],
             default: Optional[Any] = None) -> Dict[T, Any]:
    r"""Removes keys from a dictionary and returns their values.

    Args:
        dict\_ (dict): A dictionary from which items are removed.
        pop_keys: A key or a list of keys to remove and return respective
            values or :attr:`default`.
        default (optional): Value to be returned when a key is not in
            :attr:`dict_`. The default value is `None`.

    Returns:
        A ``dict`` of the items removed from :attr:`dict_`.
    """
    if not isinstance(pop_keys, (list, tuple)):
        pop_keys = cast(List[T], [pop_keys])
    ret_dict = {key: dict_.pop(key, default) for key in pop_keys}
    return ret_dict


def flatten_dict(dict_: AnyDict, parent_key: str = "", sep: str = "."):
    r"""Flattens a nested dictionary. Namedtuples within the dictionary are
    also converted to dictionaries.

    Adapted from:
    https://github.com/google/seq2seq/blob/master/seq2seq/models/model_base.py

    Args:
        dict\_ (dict): The dictionary to flatten.
        parent_key (str): A prefix to prepend to each key.
        sep (str): Separator that intervenes between parent and child keys.
            For example, if :attr:`sep` == ``"."``, then ``{ "a": { "b": 3 } }``
            is converted into ``{ "a.b": 3 }``.

    Returns:
        A new flattened ``dict``.
    """
    items: List[Tuple[str, Any]] = []
    for key, value in dict_.items():
        key_ = parent_key + sep + key if parent_key else key
        if isinstance(value, MutableMapping):
            items.extend(flatten_dict(value, key_, sep=sep).items())
        elif isinstance(value, tuple) and hasattr(value, '_asdict'):
            # namedtuple
            dict_items = collections.OrderedDict(
                zip(value._fields, value))  # type: ignore
            items.extend(flatten_dict(dict_items, key_, sep=sep).items())
        else:
            items.append((key_, value))
    return dict(items)


def default_str(str_: Optional[str], default: str) -> str:
    r"""Returns :attr:`str_` if it is not `None` or empty, otherwise returns
    :attr:`default_str`.

    Args:
        str\_: A string.
        default: A string.

    Returns:
        Either :attr:`str_` or :attr:`default_str`.
    """
    if str_ is not None and str_ != "":
        return str_
    else:
        return default


def uniquify_str(str_: str, str_set: Collection[str]) -> str:
    r"""Uniquifies :attr:`str_` if :attr:`str_` is included in :attr:`str_set`.

    This is done by appending a number to :attr:`str_`. Returns
    :attr:`str_` directly if it is not included in :attr:`str_set`.

    Args:
        str\_ (string): A string to uniquify.
        str_set (set, dict, or list): A collection of strings. The returned
            string is guaranteed to be different from the elements in the
            collection.

    Returns:
        The uniquified string. Returns :attr:`str_` directly if it is
        already unique.

    Example:

        .. code-block:: python

            print(uniquify_str('name', ['name', 'name_1']))
            # 'name_2'

    """
    if str_ not in str_set:
        return str_
    else:
        for i in range(1, len(str_set) + 1):
            unique_str = str_ + "_%d" % i
            if unique_str not in str_set:
                return unique_str
    raise ValueError("Failed to uniquify string: " + str_)


def _recur_split(s: MaybeSeq[str],
                 dtype_as: Collection[str]) -> MaybeSeq[str]:
    r"""Splits (possibly nested list of) strings recursively.
    """
    if isinstance(s, str):
        return _maybe_list_to_array(s.split(), dtype_as)
    else:
        s_ = [_recur_split(si, dtype_as) for si in s]
        return _maybe_list_to_array(s_, s)


def strip_token(str_: MaybeSeq[str], token: str,
                is_token_list: bool = False) -> MaybeSeq[str]:
    r"""Returns a copy of strings with leading and trailing tokens removed.

    Note that besides :attr:`token`, all leading and trailing whitespace
    characters are also removed.

    If :attr:`is_token_list` is False, then the function assumes tokens in
    :attr:`str_` are separated with whitespace character.

    Args:
        str\_: A ``str``, or an ``n``-D numpy array or (possibly nested)
            list of ``str``.
        token (str): The token to strip, e.g., the ``"<PAD>"`` token defined in
            :class:`~texar.torch.data.SpecialTokens`.
        is_token_list (bool): Whether each sentence in :attr:`str_` is a list
            of tokens. If False, each sentence in :attr:`str_` is assumed to
            contain tokens separated with space character.

    Returns:
        The stripped strings of the same structure/shape as :attr:`str_`.

    Example:

        .. code-block:: python

            str_ = '<PAD> a sentence <PAD> <PAD>  '
            str_stripped = strip_token(str_, '<PAD>')
            # str_stripped == 'a sentence'

            str_ = ['<PAD>', 'a', 'sentence', '<PAD>', '<PAD>', '', '']
            str_stripped = strip_token(str_, '<PAD>', is_token_list=True)
            # str_stripped == 'a sentence'
    """

    def _recur_strip(s):
        if isinstance(s, str):
            if token == "":
                return ' '.join(s.strip().split())
            else:
                return (' '.join(s.strip().split())
                        .replace(' ' + token, '').replace(token + ' ', ''))
        else:
            s_ = [_recur_strip(si) for si in s]
            return _maybe_list_to_array(s_, s)

    s = str_

    if is_token_list:
        s = str_join(s)  # type: ignore

    strp_str = _recur_strip(s)

    if is_token_list:
        strp_str = _recur_split(strp_str, str_)

    return strp_str


def strip_eos(str_: MaybeSeq[str], eos_token: str = '<EOS>',
              is_token_list: bool = False) -> MaybeSeq[str]:
    r"""Remove the EOS token and all subsequent tokens.

    If :attr:`is_token_list` is False, then the function assumes tokens in
    :attr:`str_` are separated with whitespace character.

    Args:
        str\_: A ``str``, or an ``n``-D numpy array or (possibly nested)
            list of ``str``.
        eos_token (str): The EOS token. Default is ``"<EOS>"`` as defined in
            :class:`~texar.torch.data.SpecialTokens`.EOS
        is_token_list (bool): Whether each sentence in :attr:`str_` is a list
            of tokens. If False, each sentence in :attr:`str_` is assumed to
            contain tokens separated with space character.

    Returns:
        Strings of the same structure/shape as :attr:`str_`.
    """

    def _recur_strip(s):
        if isinstance(s, str):
            s_tokens = s.split()
            if eos_token in s_tokens:
                return ' '.join(s_tokens[:s_tokens.index(eos_token)])
            else:
                return s
        else:
            s_ = [_recur_strip(si) for si in s]
            return _maybe_list_to_array(s_, s)

    s = str_

    if is_token_list:
        s = str_join(s)  # type: ignore

    strp_str = _recur_strip(s)

    if is_token_list:
        strp_str = _recur_split(strp_str, str_)

    return strp_str


_strip_eos_ = strip_eos


def strip_bos(str_: MaybeSeq[str], bos_token: str = '<BOS>',
              is_token_list: bool = False) -> MaybeSeq[str]:
    r"""Remove all leading BOS tokens.

    Note that besides :attr:`bos_token`, all leading and trailing whitespace
    characters are also removed.

    If :attr:`is_token_list` is False, then the function assumes tokens in
    :attr:`str_` are separated with whitespace character.

    Args:
        str_: A ``str``, or an ``n``-D numpy array or (possibly nested)
            list of ``str``.
        bos_token (str): The BOS token. Default is ``"<BOS>"`` as defined in
            :class:`~texar.torch.data.SpecialTokens`.BOS
        is_token_list (bool): Whether each sentence in :attr:`str_` is a list
            of tokens. If False, each sentence in :attr:`str_` is assumed to
            contain tokens separated with space character.

    Returns:
        Strings of the same structure/shape as :attr:`str_`.
    """

    def _recur_strip(s):
        if isinstance(s, str):
            if bos_token == '':
                return ' '.join(s.strip().split())
            else:
                return ' '.join(s.strip().split()).replace(bos_token + ' ', '')
        else:
            s_ = [_recur_strip(si) for si in s]
            return _maybe_list_to_array(s_, s)

    s = str_

    if is_token_list:
        s = str_join(s)  # type: ignore

    strp_str = _recur_strip(s)

    if is_token_list:
        strp_str = _recur_split(strp_str, str_)

    return strp_str


_strip_bos_ = strip_bos


# pylint: disable=redefined-outer-name

def strip_special_tokens(str_: MaybeSeq[str],
                         strip_pad: Optional[str] = '<PAD>',
                         strip_bos: Optional[str] = '<BOS>',
                         strip_eos: Optional[str] = '<EOS>',
                         is_token_list: bool = False) -> MaybeSeq[str]:
    r"""Removes special tokens in strings, including:

        - Removes EOS and all subsequent tokens
        - Removes leading and and trailing PAD tokens
        - Removes leading BOS tokens

    Note that besides the special tokens, all leading and trailing whitespace
    characters are also removed.

    This is a joint function of :func:`strip_eos`, :func:`strip_pad`, and
    :func:`strip_bos`

    Args:
        str\_: A ``str``, or an ``n``-D numpy array or (possibly nested)
            list of ``str``.
        strip_pad (str): The PAD token to strip from the strings (i.e., remove
            the leading and trailing PAD tokens of the strings). Default
            is ``"<PAD>"`` as defined in
            :class:`~texar.torch.data.SpecialTokens`.PAD.
            Set to `None` or `False` to disable the stripping.
        strip_bos (str): The BOS token to strip from the strings (i.e., remove
            the leading BOS tokens of the strings).
            Default is ``"<BOS>"`` as defined in
            :class:`~texar.torch.data.SpecialTokens`.BOS.
            Set to `None` or `False` to disable the stripping.
        strip_eos (str): The EOS token to strip from the strings (i.e., remove
            the EOS tokens and all subsequent tokens of the strings).
            Default is ``"<EOS>"`` as defined in
            :class:`~texar.torch.data.SpecialTokens`.EOS.
            Set to `None` or `False` to disable the stripping.
        is_token_list (bool): Whether each sentence in :attr:`str_` is a list
            of tokens. If `False`, each sentence in :attr:`str_` is assumed to
            contain tokens separated with space character.

    Returns:
        Strings of the same shape of :attr:`str_` with special tokens stripped.
    """
    s = str_

    if is_token_list:
        s = str_join(s)  # type: ignore

    if strip_eos is not None and strip_eos is not False:
        s = _strip_eos_(s, strip_eos, is_token_list=False)

    if strip_pad is not None and strip_pad is not False:
        s = strip_token(s, strip_pad, is_token_list=False)

    if strip_bos is not None and strip_bos is not False:
        s = _strip_bos_(s, strip_bos, is_token_list=False)

    if is_token_list:
        s = _recur_split(s, str_)

    return s


def str_join(tokens: Sequence[List], sep: str = ' ') -> Sequence[str]:
    r"""Concatenates :attr:`tokens` along the last dimension with intervening
    occurrences of :attr:`sep`.

    Args:
        tokens: An ``n``-D numpy array or (possibly nested) list of ``str``.
        sep (str): The string intervening between the tokens.

    Returns:
        An ``(n-1)``-D numpy array (or list) of ``str``.
    """

    def _recur_join(s):
        if len(s) == 0:
            return ''
        elif isinstance(s[0], str):
            return sep.join(s)
        else:
            s_ = [_recur_join(si) for si in s]
            return _maybe_list_to_array(s_, s)

    str_ = _recur_join(tokens)

    return str_


# pylint: enable=redefined-outer-name

def ceildiv(a: int, b: int) -> int:
    r"""Compute division with results rounding up.

    For example, ``5 / 2 = 2.5``, ``ceildiv(5, 2) = 3``.

    Args:
        a (int): The dividend.
        b (int): The divisor.

    Returns:
        int: The quotient, rounded up.
    """
    return -(-a // b)


def sum_tensors(xs: List[Optional[torch.Tensor]]) -> Optional[torch.Tensor]:
    r"""Sum a list of tensors with possible `None` values.

    Args:
        xs: A list of tensors.

    Returns:
        The summation of all the elements in the list.
    """
    idx = next((idx for idx, tensor in enumerate(xs) if tensor is not None), -1)
    if idx == -1:
        return None
    ret = xs[idx]
    for tensor in xs[(idx + 1):]:
        if tensor is not None:
            ret = ret + tensor
    return ret


def truncate_seq_pair(tokens_a: Union[List[int], List[str]],
                      tokens_b: Union[List[int], List[str]],
                      max_length: int):
    r"""Truncates a sequence pair in place to the maximum length.

    This is a simple heuristic which will always truncate the longer sequence
    one token at a time. This makes more sense than truncating an equal
    percent of tokens from each, since if one sequence is very short then
    each token that's truncated likely contains more information than a
    longer sequence.

    Example:
        tokens_a = [1, 2, 3, 4, 5]
        tokens_b = [6, 7]
        truncate_seq_pair(tokens_a, tokens_b, 5)
        tokens_a  # [1, 2, 3]
        tokens_b  # [6, 7]

    Args:
        tokens_a: A list of tokens or token ids.
        tokens_b: A list of tokens or token ids.
        max_length: maximum sequence length.
    """
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for various layers.
"""
import unittest

import torch
import torch.nn.functional as F
from torch import nn

from texar.torch.core import layers


class GetActivationFnTest(unittest.TestCase):
    r"""Tests :func:`texar.torch.core.layers.get_activation_fn`.
    """

    def test_get_activation_fn(self):
        r"""Tests.
        """
        fn = layers.get_activation_fn()
        self.assertEqual(fn, None)

        fn = layers.get_activation_fn('relu')
        self.assertEqual(fn, torch.relu)

        inputs = torch.randn(64, 100)

        fn = layers.get_activation_fn('leaky_relu')
        fn_output = fn(inputs)
        ref_output = F.leaky_relu(inputs)
        self.assertEqual(torch.all(torch.eq(fn_output, ref_output)), 1)

        fn = layers.get_activation_fn('leaky_relu',
                                      kwargs={'negative_slope': 0.1})
        fn_output = fn(inputs)
        ref_output = F.leaky_relu(inputs, negative_slope=0.1)
        self.assertEqual(torch.all(torch.eq(fn_output, ref_output)), 1)


class GetLayerTest(unittest.TestCase):
    r"""Tests layer creator.
    """

    def test_get_layer(self):
        r"""Tests :func:`texar.torch.core.layers.get_layer`.
        """
        hparams = {"type": "Conv1d",
                   "kwargs": {"in_channels": 16,
                              "out_channels": 32,
                              "kernel_size": 2}
                   }

        layer = layers.get_layer(hparams)
        self.assertIsInstance(layer, nn.Conv1d)

        hparams = {
            "type": "MergeLayer",
            "kwargs": {
                "layers": [
                    {"type": "Conv1d",
                     "kwargs": {"in_channels": 16,
                                "out_channels": 32,
                                "kernel_size": 2}},
                    {"type": "Conv1d",
                     "kwargs": {"in_channels": 16,
                                "out_channels": 32,
                                "kernel_size": 2}}
                ]
            }
        }
        layer = layers.get_layer(hparams)
        self.assertIsInstance(layer, layers.MergeLayer)

        hparams = {"type": "Conv1d",
                   "kwargs": {"in_channels": 16,
                              "out_channels": 32,
                              "kernel_size": 2}
                   }
        layer = layers.get_layer(hparams)
        self.assertIsInstance(layer, nn.Conv1d)

        hparams = {
            "type": nn.Conv1d(in_channels=16, out_channels=32, kernel_size=2)
        }
        layer = layers.get_layer(hparams)
        self.assertIsInstance(layer, nn.Conv1d)


class ReducePoolingLayerTest(unittest.TestCase):
    r"""Tests reduce pooling layer.
    """

    def setUp(self):
        unittest.TestCase.setUp(self)

        self._batch_size = 64
        self._emb_dim = 100
        self._seq_length = 16

    def test_max_reduce_pooling_layer(self):
        r"""Tests :class:`texar.torch.core.MaxReducePool1d`."""

        pool_layer = layers.MaxReducePool1d()
        inputs = torch.randn(self._batch_size, self._emb_dim, self._seq_length)
        output = pool_layer(inputs)
        output_reduce, _ = torch.max(inputs, dim=2)
        self.assertEqual(output.shape, torch.Size([self._batch_size,
                                                   self._emb_dim]))
        self.assertEqual(torch.all(torch.eq(output, output_reduce)), 1)

    def test_average_reduce_pooling_layer(self):
        r"""Tests :class:`texar.torch.core.AvgReducePool1d`."""

        pool_layer = layers.AvgReducePool1d()
        inputs = torch.randn(self._batch_size, self._emb_dim, self._seq_length)
        output = pool_layer(inputs)
        output_reduce = torch.mean(inputs, dim=2)
        self.assertEqual(output.shape, torch.Size([self._batch_size,
                                                   self._emb_dim]))
        self.assertEqual(torch.all(torch.eq(output, output_reduce)), 1)


class MergeLayerTest(unittest.TestCase):
    r"""Tests MergeLayer.
    """

    def test_layer_logic(self):
        r"""Test the logic of MergeLayer.
        """
        layers_ = list()
        layers_.append(nn.Conv1d(in_channels=32, out_channels=32,
                                 kernel_size=3))
        layers_.append(nn.Conv1d(in_channels=32, out_channels=32,
                                 kernel_size=3))
        layers_.append(nn.Conv1d(in_channels=32, out_channels=32,
                                 kernel_size=3))

        modes = ["concat", "sum", "mean", "prod", "max", "min", "logsumexp",
                 "elemwise_sum", "elemwise_mul"]

        for mode in modes:
            m_layer = layers.MergeLayer(layers_, mode=mode)
            input = torch.randn(32, 32, 10)
            output = m_layer(input)

            if mode == "concat":
                self.assertEqual(output.shape, torch.Size([32, 32, 24]))
            elif mode == "elemwise_sum" or mode == "elemwise_mul":
                self.assertEqual(output.shape, torch.Size([32, 32, 8]))
            else:
                self.assertEqual(output.shape, torch.Size([32, 32]))

        for mode in ["and", "or"]:
            m_layer = layers.MergeLayer(layers=None, mode=mode)
            input = torch.ones(32, 32, 10, dtype=torch.uint8)
            output = m_layer(input)

            self.assertEqual(output.shape, torch.Size([32, 32]))

    def test_empty_merge_layer(self):
        r"""Test the output of MergeLayer with empty layers.
        """
        m_layer = layers.MergeLayer(layers=None)
        input = torch.randn(32, 32, 10)
        output = m_layer(input)
        self.assertEqual(torch.all(torch.eq(output, input)), 1)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for data related operations.
"""
import copy
import tempfile
import unittest

import numpy as np

from texar.torch.data.data.data_iterators import DataIterator
from texar.torch.data.data.mono_text_data import MonoTextData
from texar.torch.data.vocabulary import SpecialTokens


class MonoTextDataTest(unittest.TestCase):
    r"""Tests text data class.
    """

    def setUp(self):
        # Create test data
        vocab_list = ['word', '词']
        vocab_file = tempfile.NamedTemporaryFile()
        vocab_file.write('\n'.join(vocab_list).encode("utf-8"))
        vocab_file.flush()
        self._vocab_file = vocab_file
        self._vocab_size = len(vocab_list)

        text = ['This is a test sentence .', '词 词 。']
        text_file = tempfile.NamedTemporaryFile()
        text_file.write('\n'.join(text).encode("utf-8"))
        text_file.flush()
        self._text_file = text_file

        self._hparams = {
            "num_epochs": 1,
            "batch_size": 3,
            "dataset": {
                "files": self._text_file.name,
                "vocab_file": self._vocab_file.name,
            }
        }

        self.upper_cased_text = []
        for sent in text:
            upper_cased_tokens = sent.upper().split(" ")
            upper_cased_tokens.insert(0, SpecialTokens.BOS)
            upper_cased_tokens.append(SpecialTokens.EOS)
            self.upper_cased_text.append(upper_cased_tokens)

        max_length = max([len(tokens) for tokens in self.upper_cased_text])
        self.upper_cased_text = [sent + [''] * (max_length - len(sent)) for sent
                                 in self.upper_cased_text]

    def _run_and_test(self,
                      hparams,
                      test_batch_size=False,
                      test_transform=False):
        # Construct database
        text_data = MonoTextData(hparams)
        self.assertEqual(text_data.vocab.size,
                         self._vocab_size + len(text_data.vocab.special_tokens))

        iterator = DataIterator(text_data)

        for data_batch in iterator:
            self.assertEqual(set(data_batch.keys()),
                             set(text_data.list_items()))

            if test_batch_size:
                self.assertEqual(len(data_batch['text']), hparams['batch_size'])

            if test_transform:
                for i in range(len(data_batch['text'])):
                    text_ = data_batch['text'][i]
                    self.assertTrue(text_ in self.upper_cased_text)

            max_seq_length = text_data.hparams.dataset.max_seq_length
            mode = text_data.hparams.dataset.length_filter_mode

            max_l = max_seq_length
            if max_seq_length is not None:
                if text_data.hparams.dataset.eos_token != '':
                    max_l += 1
                if text_data.hparams.dataset.bos_token != '':
                    max_l += 1

            if max_seq_length == 6:
                for length in data_batch['length']:
                    self.assertLessEqual(length, max_l)
                if mode == "discard":
                    for length in data_batch['length']:
                        self.assertEqual(length, 5)
                elif mode == "truncate":
                    num_length_6 = 0
                    for length in data_batch['length']:
                        num_length_6 += int(length == 6)
                    self.assertGreater(num_length_6, 0)
                else:
                    raise ValueError("Unknown mode: %s" % mode)

            if text_data.hparams.dataset.pad_to_max_seq_length:
                for x in data_batch['text']:
                    self.assertEqual(len(x), max_l)
                for x in data_batch['text_ids']:
                    self.assertEqual(len(x), max_l)

    def test_default_setting(self):
        r"""Tests the logic of MonoTextData.
        """
        self._run_and_test(self._hparams)

    def test_batching(self):
        r"""Tests different batching.
        """
        # disallow smaller final batch
        hparams = copy.deepcopy(self._hparams)
        hparams.update({"allow_smaller_final_batch": False})
        self._run_and_test(hparams, test_batch_size=True)

    @unittest.skip("bucketing is not yet implemented")
    def test_bucketing(self):
        r"""Tests bucketing.
        """
        hparams = copy.deepcopy(self._hparams)
        hparams.update({
            "bucket_boundaries": [7],
            "bucket_batch_sizes": [6, 4]})

        text_data = MonoTextData(hparams)
        iterator = DataIterator(text_data)

        hparams.update({
            "bucket_boundaries": [7],
            "bucket_batch_sizes": [7, 7],
            "allow_smaller_final_batch": False})

        text_data_1 = MonoTextData(hparams)
        iterator_1 = DataIterator(text_data_1)

        for data_batch, data_batch_1 in zip(iterator, iterator_1):
            length = data_batch['length'][0]
            if length < 7:
                last_batch_size = hparams['num_epochs'] % 6
                self.assertTrue(
                    len(data_batch['text']) == 6 or
                    len(data_batch['text']) == last_batch_size)
            else:
                last_batch_size = hparams['num_epochs'] % 4
                self.assertTrue(
                    len(data_batch['text']) == 4 or
                    len(data_batch['text']) == last_batch_size)

            self.assertEqual(len(data_batch_1['text']), 7)

    def test_shuffle(self):
        r"""Tests different shuffling strategies.
        """
        hparams = copy.deepcopy(self._hparams)
        hparams.update({
            "shard_and_shuffle": True,
            "shuffle_buffer_size": 1})
        self._run_and_test(hparams)

    def test_prefetch(self):
        r"""Tests prefetching.
        """
        hparams = copy.deepcopy(self._hparams)
        hparams.update({"prefetch_buffer_size": 2})
        self._run_and_test(hparams)

    def test_other_transformations(self):
        r"""Tests use of other transformations
        """

        upper_func = lambda x: [w.upper() for w in x]

        hparams = copy.deepcopy(self._hparams)
        hparams["dataset"].update(
            {"other_transformations": [upper_func]})
        self._run_and_test(hparams, test_transform=True)

    def test_list_items(self):
        r"""Tests the item names of the output data.
        """
        text_data = MonoTextData(self._hparams)
        self.assertSetEqual(set(text_data.list_items()),
                            {"text", "text_ids", "length"})

        hparams = copy.deepcopy(self._hparams)
        hparams["dataset"]["data_name"] = "data"
        text_data = MonoTextData(hparams)
        self.assertSetEqual(set(text_data.list_items()),
                            {"data_text", "data_text_ids", "data_length"})

    def test_length_discard(self):
        r"""Tests discard length seq.
        """
        hparams = copy.deepcopy(self._hparams)
        hparams["dataset"].update({"max_seq_length": 4,
                                   "length_filter_mode": "discard"})
        self._run_and_test(hparams)

    def test_length_truncate(self):
        r"""Tests truncation.
        """
        hparams = copy.deepcopy(self._hparams)
        hparams["dataset"].update({"max_seq_length": 4,
                                   "length_filter_mode": "truncate"})
        hparams["shuffle"] = False
        hparams["allow_smaller_final_batch"] = False
        self._run_and_test(hparams)

    def test_pad_to_max_length(self):
        r"""Tests padding.
        """
        hparams = copy.deepcopy(self._hparams)
        hparams["dataset"].update({"max_seq_length": 10,
                                   "length_filter_mode": "truncate",
                                   "pad_to_max_seq_length": True})
        self._run_and_test(hparams)


@unittest.skip("Skipping until Variable Utterance is implemented")
class VarUttMonoTextDataTest(unittest.TestCase):
    r"""Tests variable utterance text data class.
    """

    def setUp(self):
        # Create test data
        vocab_list = ['word', 'sentence', '词', 'response', 'dialog', '1', '2']
        vocab_file = tempfile.NamedTemporaryFile()
        vocab_file.write('\n'.join(vocab_list).encode("utf-8"))
        vocab_file.flush()
        self._vocab_file = vocab_file
        self._vocab_size = len(vocab_list)

        text = [
            'This is a dialog 1 sentence . ||| This is a dialog 1 sentence . '
            '||| This is yet another dialog 1 sentence .',  # //
            'This is a dialog 2 sentence . ||| '
            'This is also a dialog 2 sentence . ',  # //
            '词 词 词 ||| word',  # //
            'This This',  # //
            '1 1 1 ||| 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 ||| 1 1 1 ||| 2'
        ]
        text_file = tempfile.NamedTemporaryFile()
        text_file.write('\n'.join(text).encode("utf-8"))
        text_file.flush()
        self._text_file = text_file

        self._hparams = {
            "num_epochs": 1,
            "batch_size": 3,
            "shuffle": False,
            "dataset": {
                "files": self._text_file.name,
                "vocab_file": self._vocab_file.name,
                "variable_utterance": True,
                "max_utterance_cnt": 3,
                "max_seq_length": 10
            }
        }

    def _run_and_test(self, hparams):
        # Construct database
        text_data = MonoTextData(hparams)
        self.assertEqual(text_data.vocab.size,
                         self._vocab_size + len(text_data.vocab.special_tokens))

        iterator = DataIterator(text_data)

        for data_batch in iterator:
            # Run the logics
            self.assertEqual(set(data_batch.keys()),
                             set(text_data.list_items()))

            # Test utterance count
            utt_ind = np.sum(data_batch["text_ids"], 2) != 0
            utt_cnt = np.sum(utt_ind, 1)
            self.assertListEqual(
                data_batch[text_data.utterance_cnt_name].tolist(),
                utt_cnt.tolist())

            if text_data.hparams.dataset.pad_to_max_seq_length:
                max_l = text_data.hparams.dataset.max_seq_length
                max_l += text_data._decoder.added_length
                for x in data_batch['text']:
                    for xx in x:
                        self.assertEqual(len(xx), max_l)
                for x in data_batch['text_ids']:
                    for xx in x:
                        self.assertEqual(len(xx), max_l)

    def test_default_setting(self):
        r"""Tests the logics of the text data.
        """
        self._run_and_test(self._hparams)

    def test_pad_to_max_length(self):
        r"""Tests padding.
        """
        hparams = copy.copy(self._hparams)
        hparams["dataset"].update({"max_seq_length": 20,
                                   "length_filter_mode": "truncate",
                                   "pad_to_max_seq_length": True})
        self._run_and_test(hparams)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for data related operations.
"""
import torch

import copy
import tempfile
import numpy as np

import unittest

from texar.torch.data import DataIterator, ScalarData
from texar.torch.utils.dtypes import torch_bool


class ScalarDataTest(unittest.TestCase):
    """Tests scalar data class.
    """

    def setUp(self):
        # Create test data
        int_data = np.linspace(0, 100, num=101, dtype=np.int32).tolist()
        int_data = [str(i) for i in int_data]
        int_file = tempfile.NamedTemporaryFile()
        int_file.write('\n'.join(int_data).encode("utf-8"))
        int_file.flush()
        self._int_file = int_file

        self._int_hparams = {
            "num_epochs": 1,
            "batch_size": 1,
            "shuffle": False,
            "dataset": {
                "files": self._int_file.name,
                "data_type": "int",
                "data_name": "label"
            }
        }

        self._float_hparams = {
            "num_epochs": 1,
            "batch_size": 1,
            "shuffle": False,
            "dataset": {
                "files": self._int_file.name,
                "data_type": "float",
                "data_name": "feat"
            }
        }

        bool_data = [0, 1]
        bool_data = [str(i) for i in bool_data]
        bool_file = tempfile.NamedTemporaryFile()
        bool_file.write('\n'.join(bool_data).encode("utf-8"))
        bool_file.flush()
        self._bool_file = bool_file

        self._bool_hparams = {
            "num_epochs": 1,
            "batch_size": 1,
            "shuffle": False,
            "dataset": {
                "files": self._bool_file.name,
                "data_type": "bool",
                "data_name": "feat"
            }
        }

    def _run_and_test(self, hparams, test_transform=False):
        # Construct database
        scalar_data = ScalarData(hparams)

        self.assertEqual(scalar_data.list_items()[0],
                         hparams["dataset"]["data_name"])

        iterator = DataIterator(scalar_data)

        i = 0
        for batch in iterator:
            self.assertEqual(set(batch.keys()),
                             set(scalar_data.list_items()))
            value = batch[scalar_data.data_name][0]
            if test_transform:
                self.assertEqual(2 * i, value)
            else:
                self.assertEqual(i, value)
            i += 1
            data_type = hparams["dataset"]["data_type"]
            if data_type == "int":
                self.assertEqual(value.dtype, torch.int32)
            elif data_type == "float":
                self.assertEqual(value.dtype, torch.float32)
            elif data_type == "bool":
                self.assertTrue(value.dtype, torch_bool)
            self.assertIsInstance(value, torch.Tensor)

    def test_default_setting(self):
        """Tests the logic of ScalarData.
        """
        self._run_and_test(self._int_hparams)
        self._run_and_test(self._float_hparams)
        self._run_and_test(self._bool_hparams)

    def test_shuffle(self):
        """Tests results of toggling shuffle.
        """
        hparams = copy.copy(self._int_hparams)
        hparams["batch_size"] = 10
        scalar_data = ScalarData(hparams)
        iterator = DataIterator(scalar_data)

        hparams_sfl = copy.copy(hparams)
        hparams_sfl["shuffle"] = True
        scalar_data_sfl = ScalarData(hparams_sfl)
        iterator_sfl = DataIterator(scalar_data_sfl)

        vals = []
        vals_sfl = []

        for batch, batch_sfl in zip(iterator, iterator_sfl):
            vals += batch["label"].tolist()
            vals_sfl += batch_sfl["label"].tolist()

        self.assertEqual(len(vals), len(vals_sfl))
        self.assertSetEqual(set(vals), set(vals_sfl))

    def test_transform(self):
        """Tests transform logic.
        """
        hparams = copy.deepcopy(self._int_hparams)
        hparams["dataset"].update(
            {"other_transformations": [lambda x: x * 2]})
        self._run_and_test(hparams, test_transform=True)

        hparams = copy.deepcopy(self._float_hparams)
        hparams["dataset"].update(
            {"other_transformations": [lambda x: x * 2]})
        self._run_and_test(hparams, test_transform=True)

    def test_unsupported_scalar_types(self):
        """Tests exception for unsupported scalar types.
        """
        hparams = copy.copy(self._int_hparams)
        hparams["dataset"].update({
            "data_type": "XYZ"
        })

        with self.assertRaises(ValueError):
            self._run_and_test(hparams)

        hparams = copy.copy(self._int_hparams)
        hparams["dataset"].update({
            "data_type": "str"
        })

        with self.assertRaises(ValueError):
            self._run_and_test(hparams)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for data related operations.
"""
import copy
import os
import tempfile
import unittest

import numpy as np
import torch

from texar.torch.data.data.data_iterators import DataIterator
from texar.torch.data.data.multi_aligned_data import MultiAlignedData
from texar.torch.data.data.record_data import RecordData


class MultiAlignedDataTest(unittest.TestCase):
    """Tests multi aligned text data class.
    """

    def setUp(self):
        # Create test data
        vocab_list = ['This', 'is', 'a', 'word', '词']
        vocab_file = tempfile.NamedTemporaryFile()
        vocab_file.write('\n'.join(vocab_list).encode("utf-8"))
        vocab_file.flush()
        self._vocab_file = vocab_file
        self._vocab_size = len(vocab_list)

        text_0 = ['This is a sentence from source .', '词 词 。 source']
        text_0_file = tempfile.NamedTemporaryFile()
        text_0_file.write('\n'.join(text_0).encode("utf-8"))
        text_0_file.flush()
        self._text_0_file = text_0_file

        text_1 = ['This is a sentence from target .', '词 词 。 target']
        text_1_file = tempfile.NamedTemporaryFile()
        text_1_file.write('\n'.join(text_1).encode("utf-8"))
        text_1_file.flush()
        self._text_1_file = text_1_file

        text_2 = [
            'This is a sentence from dialog . ||| dialog ',
            '词 词 。 ||| 词 dialog']
        text_2_file = tempfile.NamedTemporaryFile()
        text_2_file.write('\n'.join(text_2).encode("utf-8"))
        text_2_file.flush()
        self._text_2_file = text_2_file

        int_3 = [0, 1]
        int_3_file = tempfile.NamedTemporaryFile()
        int_3_file.write(('\n'.join([str(_) for _ in int_3])).encode("utf-8"))
        int_3_file.flush()
        self._int_3_file = int_3_file

        self._record_filepath = os.path.join(
            tempfile.mkdtemp(), 'test.pkl')
        self._feature_types = {
            'number1': ('tf.int64', 'stacked_tensor'),
            'number2': ('tf.int64', 'stacked_tensor'),
            'text': ('tf.string', 'stacked_tensor')
        }

        features = [{
            "number1": 128,
            "number2": 512,
            "text": "This is a sentence for TFRecord 词 词 。"
        },
            {
                "number1": 128,
                "number2": 512,
                "text": "This is a another sentence for TFRecord 词 词 。"
            }]
        # Prepare Validation data
        with RecordData.writer(self._record_filepath,
                               self._feature_types) as writer:
            for feature in features:
                writer.write(feature)

        # Construct database
        self._hparams = {
            "num_epochs": 1,
            "batch_size": 1,
            "datasets": [
                {  # dataset 0
                    "files": [self._text_0_file.name],
                    "vocab_file": self._vocab_file.name,
                    "bos_token": "",
                    "data_name": "0"
                },
                {  # dataset 1
                    "files": [self._text_1_file.name],
                    "vocab_share_with": 0,
                    "eos_token": "<TARGET_EOS>",
                    "data_name": "1"
                },
                {  # dataset 2
                    "files": [self._text_2_file.name],
                    "vocab_file": self._vocab_file.name,
                    "processing_share_with": 0,
                    # TODO(avinash) - Add it back once feature is added
                    "variable_utterance": False,
                    "data_name": "2"
                },
                {  # dataset 3
                    "files": self._int_3_file.name,
                    "data_type": "int",
                    "data_name": "label"
                },
                {  # dataset 4
                    "files": self._record_filepath,
                    "feature_types": self._feature_types,
                    "feature_convert_types": {
                        'number2': 'tf.float32',
                    },
                    "num_shards": 2,
                    "shard_id": 1,
                    "data_type": "record",
                    "data_name": "4"
                }
            ]
        }

    def _run_and_test(self, hparams, discard_index=None):
        # Construct database
        text_data = MultiAlignedData(hparams)
        self.assertEqual(
            text_data.vocab(0).size,
            self._vocab_size + len(text_data.vocab(0).special_tokens))

        iterator = DataIterator(text_data)
        for batch in iterator:
            self.assertEqual(set(batch.keys()),
                             set(text_data.list_items()))
            text_0 = batch['0_text']
            text_1 = batch['1_text']
            text_2 = batch['2_text']
            int_3 = batch['label']
            number_1 = batch['4_number1']
            number_2 = batch['4_number2']
            text_4 = batch['4_text']

            for t0, t1, t2, i3, n1, n2, t4 in zip(
                    text_0, text_1, text_2, int_3,
                    number_1, number_2, text_4):

                np.testing.assert_array_equal(t0[:2], t1[1:3])
                np.testing.assert_array_equal(t0[:3], t2[1:4])
                if t0[0].startswith('This'):
                    self.assertEqual(i3, 0)
                else:
                    self.assertEqual(i3, 1)
                self.assertEqual(n1, 128)
                self.assertEqual(n2, 512)
                self.assertIsInstance(n1, torch.Tensor)
                self.assertIsInstance(n2, torch.Tensor)
                self.assertIsInstance(t4, str)

            if discard_index is not None:
                hpms = text_data._hparams.datasets[discard_index]
                max_l = hpms.max_seq_length
                max_l += sum(int(x is not None and x != '')
                             for x in [text_data.vocab(discard_index).bos_token,
                                       text_data.vocab(discard_index).eos_token]
                             )
                for i in range(2):
                    for length in batch[text_data.length_name(i)]:
                        self.assertLessEqual(length, max_l)

                # TODO(avinash): Add this back once variable utterance is added
                # for lengths in batch[text_data.length_name(2)]:
                #    for length in lengths:
                #        self.assertLessEqual(length, max_l)

            for i, hpms in enumerate(text_data._hparams.datasets):
                if hpms.data_type != "text":
                    continue
                max_l = hpms.max_seq_length
                mode = hpms.length_filter_mode
                if max_l is not None and mode == "truncate":
                    max_l += sum(int(x is not None and x != '')
                                 for x in [text_data.vocab(i).bos_token,
                                           text_data.vocab(i).eos_token])
                    for length in batch[text_data.length_name(i)]:
                        self.assertLessEqual(length, max_l)

    def test_default_setting(self):
        """Tests the logics of the text data.
        """
        self._run_and_test(self._hparams)

    def test_length_filter(self):
        """Tests filtering by length.
        """
        hparams = copy.copy(self._hparams)
        hparams["datasets"][0].update(
            {"max_seq_length": 4,
             "length_filter_mode": "discard"})
        hparams["datasets"][1].update(
            {"max_seq_length": 2,
             "length_filter_mode": "truncate"})
        self._run_and_test(hparams, discard_index=0)

    def test_supported_scalar_types(self):
        """Tests scalar types supported in MultiAlignedData."""
        # int64 type
        hparams = copy.copy(self._hparams)
        hparams["datasets"][3].update({
            "data_type": "int64"
        })
        self._run_and_test(hparams)

        # float type
        hparams = copy.copy(self._hparams)
        hparams["datasets"][3].update({
            "data_type": "float"
        })
        self._run_and_test(hparams)

        # float64 type
        hparams = copy.copy(self._hparams)
        hparams["datasets"][3].update({
            "data_type": "float64"
        })
        self._run_and_test(hparams)

        # bool type
        hparams = copy.copy(self._hparams)
        hparams["datasets"][3].update({
            "data_type": "bool"
        })
        self._run_and_test(hparams)

    def test_unsupported_scalar_types(self):
        """Tests if exception is thrown for unsupported types."""
        hparams = copy.copy(self._hparams)
        hparams["datasets"][3].update({
            "data_type": "XYZ"
        })

        with self.assertRaises(ValueError):
            self._run_and_test(hparams)

        hparams = copy.copy(self._hparams)
        hparams["datasets"][3].update({
            "data_type": "str"
        })

        with self.assertRaises(ValueError):
            self._run_and_test(hparams)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for mle losses.
"""

import unittest

import torch
import torch.nn.functional as F

from texar.torch.losses import mle_losses
from texar.torch.utils.shapes import get_rank


class MLELossesTest(unittest.TestCase):
    """Tests mle losses.
    """

    def setUp(self):
        self._batch_size = 64
        self._max_time = 16
        self._num_classes = 100
        self._labels = torch.randint(
            self._num_classes, (self._batch_size, self._max_time),
            dtype=torch.int64)
        one_hot = torch.eye(self._num_classes)
        one_hot_labels = F.embedding(self._labels, one_hot)
        self._one_hot_labels = torch.reshape(
            one_hot_labels, [self._batch_size, self._max_time, -1])
        self._logits = torch.rand(
            self._batch_size, self._max_time, self._num_classes)
        self._sequence_length = torch.randint(
            high=self._max_time, size=(self._batch_size,), dtype=torch.int64)

    def _test_sequence_loss(self, loss_fn, labels, logits, sequence_length):
        loss = loss_fn(labels, logits, sequence_length)
        rank = get_rank(loss)
        self.assertEqual(rank, 0)

        loss = loss_fn(labels, logits, sequence_length,
                       sum_over_timesteps=False)
        rank = get_rank(loss)
        self.assertEqual(rank, 1)
        self.assertEqual(loss.shape, torch.Size([self._max_time]))

        loss = loss_fn(
            labels, logits, sequence_length, sum_over_timesteps=False,
            average_across_timesteps=True, average_across_batch=False)
        rank = get_rank(loss)
        self.assertEqual(rank, 1)
        self.assertEqual(loss.shape, torch.Size([self._batch_size]))

        loss = loss_fn(
            labels, logits, sequence_length, sum_over_timesteps=False,
            average_across_batch=False)
        rank = get_rank(loss)
        self.assertEqual(rank, 2)
        self.assertEqual(loss.shape, torch.Size([self._batch_size,
                                                 self._max_time]))

        sequence_length_time = torch.randint(size=(self._max_time,),
                                             high=self._batch_size)
        loss = loss_fn(
            labels, logits, sequence_length_time, sum_over_timesteps=False,
            average_across_batch=False, time_major=True)
        self.assertEqual(loss.shape, torch.Size([self._batch_size,
                                                 self._max_time]))

    def test_sequence_softmax_cross_entropy(self):
        """Tests `sequence_softmax_cross_entropy`
        """
        self._test_sequence_loss(
            mle_losses.sequence_softmax_cross_entropy,
            self._one_hot_labels, self._logits, self._sequence_length)

    def test_sequence_sparse_softmax_cross_entropy(self):
        """Tests `sequence_sparse_softmax_cross_entropy`
        """
        self._test_sequence_loss(
            mle_losses.sequence_sparse_softmax_cross_entropy,
            self._labels, self._logits, self._sequence_length)

    def test_sequence_sigmoid_cross_entropy(self):
        """Tests `texar.torch.losses.sequence_sigmoid_cross_entropy`.
        """
        self._test_sequence_loss(
            mle_losses.sequence_sigmoid_cross_entropy,
            self._one_hot_labels, self._logits, self._sequence_length)

        self._test_sequence_loss(
            mle_losses.sequence_sigmoid_cross_entropy,
            self._one_hot_labels[:, :, 0],
            self._logits[:, :, 0],
            self._sequence_length)

        loss = mle_losses.sequence_sigmoid_cross_entropy(
            logits=self._logits[:, :, 0],
            labels=torch.ones([self._batch_size, self._max_time]),
            sequence_length=self._sequence_length)
        rank = get_rank(loss)
        self.assertEqual(rank, 0)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for pre-trained XLNet tokenizer.
"""

import unittest

import os
import pickle
import tempfile

from texar.torch.data.data_utils import maybe_download
from texar.torch.data.tokenizers.xlnet_tokenizer import \
    XLNetTokenizer, SPIECE_UNDERLINE
from texar.torch.utils.test import pretrained_test


class XLNetTokenizerTest(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.SAMPLE_VOCAB = maybe_download(
            'https://github.com/huggingface/transformers/blob/master/'
            'tests/fixtures/test_sentencepiece.model?raw=true',
            self.tmp_dir.name)

        self.tokenizer = XLNetTokenizer.load(
            self.SAMPLE_VOCAB, configs={'keep_accents': True})
        self.tokenizer.save(self.tmp_dir.name)

    def tearDown(self):
        self.tmp_dir.cleanup()

    @pretrained_test
    def test_model_loading(self):
        for pretrained_model_name in \
                XLNetTokenizer.available_checkpoints():
            tokenizer = XLNetTokenizer(
                pretrained_model_name=pretrained_model_name)
            _ = tokenizer.map_text_to_token(u"This is a test")

    def test_tokenize(self):
        tokens = self.tokenizer.map_text_to_token(u'This is a test')
        self.assertListEqual(tokens, [u'▁This', u'▁is', u'▁a', u'▁t', u'est'])

        self.assertListEqual(
            self.tokenizer.map_token_to_id(tokens),
            [285, 46, 10, 170, 382])

        tokens = self.tokenizer.map_text_to_token(
            u"I was born in 92000, and this is falsé.")
        self.assertListEqual(tokens, [SPIECE_UNDERLINE + u'I',
                                      SPIECE_UNDERLINE + u'was',
                                      SPIECE_UNDERLINE + u'b',
                                      u'or', u'n', SPIECE_UNDERLINE + u'in',
                                      SPIECE_UNDERLINE + u'',
                                      u'9', u'2', u'0', u'0', u'0', u',',
                                      SPIECE_UNDERLINE + u'and',
                                      SPIECE_UNDERLINE + u'this',
                                      SPIECE_UNDERLINE + u'is',
                                      SPIECE_UNDERLINE + u'f', u'al', u's',
                                      u'é', u'.'])
        ids = self.tokenizer.map_token_to_id(tokens)
        self.assertListEqual(
            ids, [8, 21, 84, 55, 24, 19, 7, 0,
                  602, 347, 347, 347, 3, 12, 66,
                  46, 72, 80, 6, 0, 4])

        back_tokens = self.tokenizer.map_id_to_token(ids)
        self.assertListEqual(back_tokens, [SPIECE_UNDERLINE + u'I',
                                           SPIECE_UNDERLINE + u'was',
                                           SPIECE_UNDERLINE + u'b',
                                           u'or', u'n',
                                           SPIECE_UNDERLINE + u'in',
                                           SPIECE_UNDERLINE + u'', u'<unk>',
                                           u'2', u'0', u'0', u'0', u',',
                                           SPIECE_UNDERLINE + u'and',
                                           SPIECE_UNDERLINE + u'this',
                                           SPIECE_UNDERLINE + u'is',
                                           SPIECE_UNDERLINE + u'f', u'al', u's',
                                           u'<unk>', u'.'])

    def test_pickle(self):
        tokenizer = XLNetTokenizer.load(self.tmp_dir.name)
        self.assertIsNotNone(tokenizer)

        text = u"Munich and Berlin are nice cities"
        subwords = tokenizer.map_text_to_token(text)

        with tempfile.TemporaryDirectory() as tmpdirname:
            filename = os.path.join(tmpdirname, u"tokenizer.bin")
            with open(filename, "wb") as f:
                pickle.dump(tokenizer, f)
            with open(filename, "rb") as f:
                tokenizer_new = pickle.load(f)

        subwords_loaded = tokenizer_new.map_text_to_token(text)

        self.assertListEqual(subwords, subwords_loaded)

    def test_save_load(self):
        tokenizer = XLNetTokenizer.load(self.tmp_dir.name)

        before_tokens = tokenizer.map_text_to_id(
            u"He is very happy, UNwant\u00E9d,running")

        with tempfile.TemporaryDirectory() as tmpdirname:
            tokenizer.save(tmpdirname)
            tokenizer = tokenizer.load(tmpdirname)

        after_tokens = tokenizer.map_text_to_id(
            u"He is very happy, UNwant\u00E9d,running")
        self.assertListEqual(before_tokens, after_tokens)

    def test_pretrained_model_list(self):
        model_list_1 = list(XLNetTokenizer._MODEL2URL.keys())
        model_list_2 = list(XLNetTokenizer._MAX_INPUT_SIZE.keys())

        self.assertListEqual(model_list_1, model_list_2)

    def test_encode_decode(self):
        tokenizer = XLNetTokenizer.load(self.tmp_dir.name)

        input_text = u"This is a test"
        output_text = u"This is a test"

        tokens = tokenizer.map_text_to_token(input_text)
        ids = tokenizer.map_token_to_id(tokens)
        ids_2 = tokenizer.map_text_to_id(input_text)
        self.assertListEqual(ids, ids_2)

        tokens_2 = tokenizer.map_id_to_token(ids)
        text_2 = tokenizer.map_id_to_text(ids)

        self.assertEqual(text_2, output_text)

        self.assertNotEqual(len(tokens_2), 0)
        self.assertIsInstance(text_2, str)

    def test_add_tokens(self):
        tokenizer = XLNetTokenizer.load(self.tmp_dir.name)

        vocab_size = tokenizer.vocab_size
        all_size = len(tokenizer)

        self.assertNotEqual(vocab_size, 0)
        self.assertEqual(vocab_size, all_size)

        new_toks = ["aaaaabbbbbb", "cccccccccdddddddd"]
        added_toks = tokenizer.add_tokens(new_toks)
        vocab_size_2 = tokenizer.vocab_size
        all_size_2 = len(tokenizer)

        self.assertNotEqual(vocab_size_2, 0)
        self.assertEqual(vocab_size, vocab_size_2)
        self.assertEqual(added_toks, len(new_toks))
        self.assertEqual(all_size_2, all_size + len(new_toks))

        tokens = tokenizer.map_text_to_id("aaaaabbbbbb low cccccccccdddddddd l")
        self.assertGreaterEqual(len(tokens), 4)
        self.assertGreater(tokens[0], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[-2], tokenizer.vocab_size - 1)

        new_toks_2 = {'eos_token': ">>>>|||<||<<|<<",
                      'pad_token': "<<<<<|||>|>>>>|>"}
        added_toks_2 = tokenizer.add_special_tokens(new_toks_2)
        vocab_size_3 = tokenizer.vocab_size
        all_size_3 = len(tokenizer)

        self.assertNotEqual(vocab_size_3, 0)
        self.assertEqual(vocab_size, vocab_size_3)
        self.assertEqual(added_toks_2, len(new_toks_2))
        self.assertEqual(all_size_3, all_size_2 + len(new_toks_2))

        tokens = tokenizer.map_text_to_id(
            ">>>>|||<||<<|<< aaaaabbbbbb low cccccccccdddddddd "
            "<<<<<|||>|>>>>|> l")

        self.assertGreaterEqual(len(tokens), 6)
        self.assertGreater(tokens[0], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[0], tokens[1])
        self.assertGreater(tokens[-2], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[-2], tokens[-3])
        self.assertEqual(tokens[0],
                         tokenizer.map_token_to_id(tokenizer.eos_token))
        self.assertEqual(tokens[-2],
                         tokenizer.map_token_to_id(tokenizer.pad_token))

    def test_tokenizer_lower(self):
        tokenizer = XLNetTokenizer.load(
            self.SAMPLE_VOCAB, configs={'do_lower_case': True,
                                        'keep_accents': False})
        tokens = tokenizer.map_text_to_token(
            u"I was born in 92000, and this is falsé.")
        self.assertListEqual(tokens, [SPIECE_UNDERLINE + u'', u'i',
                                      SPIECE_UNDERLINE + u'was',
                                      SPIECE_UNDERLINE + u'b',
                                      u'or', u'n', SPIECE_UNDERLINE + u'in',
                                      SPIECE_UNDERLINE + u'',
                                      u'9', u'2', u'0', u'0', u'0', u',',
                                      SPIECE_UNDERLINE + u'and',
                                      SPIECE_UNDERLINE + u'this',
                                      SPIECE_UNDERLINE + u'is',
                                      SPIECE_UNDERLINE + u'f', u'al', u'se',
                                      u'.'])
        self.assertListEqual(tokenizer.map_text_to_token(u"H\u00E9llo"),
                             [u"▁he", u"ll", u"o"])

    def test_tokenizer_no_lower(self):
        tokenizer = XLNetTokenizer.load(
            self.SAMPLE_VOCAB, configs={'do_lower_case': False,
                                        'keep_accents': False})
        tokens = tokenizer.map_text_to_token(
            u"I was born in 92000, and this is falsé.")
        self.assertListEqual(tokens, [SPIECE_UNDERLINE + u'I',
                                      SPIECE_UNDERLINE + u'was',
                                      SPIECE_UNDERLINE + u'b', u'or',
                                      u'n', SPIECE_UNDERLINE + u'in',
                                      SPIECE_UNDERLINE + u'',
                                      u'9', u'2', u'0', u'0', u'0', u',',
                                      SPIECE_UNDERLINE + u'and',
                                      SPIECE_UNDERLINE + u'this',
                                      SPIECE_UNDERLINE + u'is',
                                      SPIECE_UNDERLINE + u'f', u'al', u'se',
                                      u'.'])

    def test_encode_text(self):
        text_1 = u"He is very happy"
        text_2 = u"unwanted, running"

        text_1_ids = self.tokenizer.map_text_to_id(text_1)
        text_2_ids = self.tokenizer.map_text_to_id(text_2)

        cls_token_id = self.tokenizer.map_token_to_id(self.tokenizer.cls_token)
        sep_token_id = self.tokenizer.map_token_to_id(self.tokenizer.sep_token)

        input_ids, segment_ids, input_mask = \
            self.tokenizer.encode_text(text_1, None, 4)

        self.assertListEqual(input_ids,
                             text_1_ids[:2] + [sep_token_id] + [cls_token_id])
        self.assertListEqual(segment_ids, [0, 0, 0, 2])
        self.assertListEqual(input_mask, [0, 0, 0, 0])

        input_ids, segment_ids, input_mask = \
            self.tokenizer.encode_text(text_1, text_2, 7)

        self.assertListEqual(input_ids, text_1_ids[:2] +
                             [sep_token_id] + text_2_ids[:2] + [sep_token_id] +
                             [cls_token_id])
        self.assertListEqual(segment_ids, [0, 0, 0, 1, 1, 1, 2])
        self.assertListEqual(input_mask, [0, 0, 0, 0, 0, 0, 0])

    def test_encode_text_for_generation(self):
        text_1 = u"lower newer"

        text_1_ids = self.tokenizer.map_text_to_id(text_1)

        input_ids, seq_len = \
            self.tokenizer.encode_text_for_generation(text=text_1,
                                                      max_seq_length=10)

        bos_token_id = self.tokenizer.map_token_to_id(self.tokenizer.bos_token)
        eos_token_id = self.tokenizer.map_token_to_id(self.tokenizer.eos_token)
        pad_token_id = self.tokenizer.map_token_to_id(self.tokenizer.pad_token)

        self.assertListEqual(input_ids,
                             [bos_token_id] + text_1_ids + [eos_token_id] +
                             [pad_token_id, pad_token_id, pad_token_id])
        self.assertEqual(seq_len, 7)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Type annotation helpers.
"""
import os
from typing import Dict, List, Sequence, Tuple, TypeVar, Union

__all__ = [
    'MaybeTuple',
    'MaybeList',
    'MaybeSeq',
    'MaybeDict',
    'PathLike',
]

T = TypeVar('T')
MaybeTuple = Union[T, Tuple[T, ...]]
MaybeList = Union[T, List[T]]
MaybeSeq = Union[T, Sequence[T]]
MaybeDict = Union[T, Dict[str, T]]
PathLike = TypeVar('PathLike', str, os.PathLike)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The BLEU metric.
"""

import logging
import os
import re
import shutil
import subprocess
import tempfile
from typing import List

import numpy as np

from texar.torch.utils.dtypes import compat_as_text
from texar.torch.utils.types import MaybeList

__all__ = [
    "sentence_bleu_moses",
    "corpus_bleu_moses",
]


def _maybe_list_to_str(list_or_str: MaybeList[str]) -> str:
    if isinstance(list_or_str, (tuple, list, np.ndarray)):
        return ' '.join(list_or_str)
    return list_or_str


def _parse_multi_bleu_ret(bleu_str: str,
                          return_all: bool = False) -> MaybeList[float]:
    bleu_score = re.search(r"BLEU = (.+?),", bleu_str).group(1)  # type: ignore
    bleu_score = np.float32(bleu_score)

    if return_all:
        bleus = re.search(r", (.+?)/(.+?)/(.+?)/(.+?) ", bleu_str)
        bleus = [bleus.group(group_idx)  # type: ignore
                 for group_idx in range(1, 5)]
        bleus = [np.float32(b) for b in bleus]
        bleu_score = [bleu_score] + bleus

    return bleu_score


def sentence_bleu_moses(references: List[MaybeList[str]],
                        hypothesis: MaybeList[str],
                        lowercase: bool = False,
                        return_all: bool = False) -> MaybeList[float]:
    r"""Calculates BLEU score of a hypothesis sentence using the
    **MOSES `multi-bleu.perl`** script.

    Args:
        references: A list of reference for the hypothesis.
            Each reference can be either a string, or a list of string tokens.
            List can also be numpy array.
        hypothesis: A hypothesis sentence.
            The hypothesis can be either a string, or a list of string tokens.
            List can also be numpy array.
        lowercase (bool): If `True`, pass the ``"-lc"`` flag to the
            `multi-bleu` script.
        return_all (bool): If `True`, returns BLEU and all n-gram
            precisions.

    Returns:
        If :attr:`return_all` is `False` (default), returns a ``float32``
        BLEU score.

        If :attr:`return_all` is `True`, returns a list of 5 ``float32``
        scores: ``[BLEU, 1-gram precision, ..., 4-gram precision]``.
    """
    return corpus_bleu_moses(
        [references], [hypothesis], lowercase=lowercase, return_all=return_all)


def corpus_bleu_moses(list_of_references: List[List[MaybeList[str]]],
                      hypotheses: List[MaybeList[str]],
                      lowercase: bool = False,
                      return_all: bool = False) -> MaybeList[float]:
    r"""Calculates corpus-level BLEU score using the
    **MOSES `multi-bleu.perl`** script.

    Args:
        list_of_references: A list of lists of references for each hypothesis.
            Each reference can be either a string, or a list of string tokens.
            List can also be numpy array.
        hypotheses: A list of hypothesis sentences.
            Each hypothesis can be either a string, or a list of string tokens.
            List can also be numpy array.
        lowercase (bool): If `True`, pass the ``"-lc"`` flag to the
            `multi-bleu` script.
        return_all (bool): If `True`, returns BLEU and all
            n-gram precisions.

    Returns:
        If :attr:`return_all` is `False` (default), returns a ``float32``
        BLEU score.

        If :attr:`return_all` is `True`, returns a list of 5 ``float32``
        scores: ``[BLEU, 1-gram precision, ..., 4-gram precision]``.
    """
    list_of_references = compat_as_text(list_of_references)
    hypotheses = compat_as_text(hypotheses)

    if np.size(hypotheses) == 0:
        return np.float32(0.)

    # Get multi-bleu.perl
    cur_dir = os.path.dirname(os.path.realpath(__file__))
    multi_bleu_path = os.path.abspath(
        os.path.join(cur_dir, "..", "..", "..",
                     "bin", "utils", "multi-bleu.perl"))

    # Create a temporary folder containing hypothesis and reference files
    result_path = tempfile.mkdtemp()
    # Create hypothesis file
    hfile_path = os.path.join(result_path, 'hyp')
    hyps = [_maybe_list_to_str(h) for h in hypotheses]
    with open(hfile_path, 'w', encoding='utf-8') as hfile:
        text = "\n".join(hyps)
        hfile.write(text)
        hfile.write("\n")
    # Create reference files
    max_nrefs = max([len(refs) for refs in list_of_references])
    rfile_path = os.path.join(result_path, 'ref')
    for rid in range(max_nrefs):
        with open(rfile_path + str(rid), 'w', encoding='utf-8') as rfile:
            for refs in list_of_references:
                if rid < len(refs):
                    ref = _maybe_list_to_str(refs[rid])
                    rfile.write(ref + "\n")
                else:
                    rfile.write("\n")

    # Calculate BLEU
    multi_bleu_cmd = [multi_bleu_path]
    if lowercase:
        multi_bleu_cmd += ["-lc"]
    multi_bleu_cmd += [rfile_path]
    with open(hfile_path, "r") as hyp_input:
        try:
            multi_bleu_ret = subprocess.check_output(
                multi_bleu_cmd, stdin=hyp_input, stderr=subprocess.STDOUT)
            multi_bleu_ret_ = multi_bleu_ret.decode("utf-8")
            bleu_score = _parse_multi_bleu_ret(multi_bleu_ret_, return_all)
        except subprocess.CalledProcessError as error:
            if error.output is not None:
                logging.warning("multi-bleu.perl returned non-zero exit code")
                logging.warning(error.output)
            if return_all:
                bleu_score = [np.float32(0.0)] * 5
            else:
                bleu_score = np.float32(0.0)

    shutil.rmtree(result_path)

    return np.float32(bleu_score)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for Transformer decoder.
"""
import unittest

import torch

from texar.torch.modules.decoders.decoder_helpers import (
    GreedyEmbeddingHelper, TopKSampleEmbeddingHelper, TopPSampleEmbeddingHelper)


class SamplerTest(unittest.TestCase):
    r"""Tests decoder helper utilities.
    """

    def setUp(self):
        self.logits = torch.Tensor([[0.2, 0.2, 0.55, 0.05]])
        # softmax values for above tensor is
        # tensor([[0.2337, 0.2337, 0.3316, 0.2011]])
        self.start_token = torch.LongTensor([1])
        self.end_token = 2

    def test_greedy_sampler(self):
        """Tests Greedy Sampler."""
        sampler = GreedyEmbeddingHelper(start_tokens=self.start_token,
                                        end_token=self.end_token)
        index = sampler.sample(time=0, outputs=self.logits)
        assert torch.equal(index, torch.argmax(self.logits, dim=1))

    def test_top_k_sampler(self):
        """Tests Top-K Sampler."""
        sampler = TopKSampleEmbeddingHelper(start_tokens=self.start_token,
                                            end_token=self.end_token, top_k=1)
        index = sampler.sample(time=0, outputs=self.logits)
        assert torch.equal(index, torch.argmax(self.logits, dim=1))

    def test_top_p_sampler(self):
        """Tests Top-P Sampler also known as Nucleus Sampler."""
        sampler = TopPSampleEmbeddingHelper(start_tokens=self.start_token,
                                            end_token=self.end_token, p=0.6)
        index = sampler.sample(time=0, outputs=self.logits)
        assert index.item() in [0, 1, 2]


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for generation related operations.
"""
import unittest

from texar.torch.run.metric.generation import *
from texar.torch.evals.bleu import corpus_bleu


class GenerationMetricTest(unittest.TestCase):
    def setUp(self) -> None:
        self.hypotheses = [
            "this is a test sentence to evaluate the good bleu score . 词",
            "i believe that that the script is 词 perfectly correct .",
            "this score should be pretty bad".split(),
        ]
        self.references = [
            "this is a test sentence to evaluate the good score .",
            "i believe that the script is perfectly correct .".split(),
            "yeah , this is a totally different sentence .",
        ]

    def test_bleu(self):
        metric = BLEU(pred_name="", label_name="")
        for idx, (hyp, ref) in enumerate(zip(self.hypotheses, self.references)):
            metric.add([hyp], [ref])
            value = metric.value()
            answer = corpus_bleu([[r] for r in self.references[:(idx + 1)]],
                                 self.hypotheses[:(idx + 1)])
            self.assertAlmostEqual(value, answer)
